from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS

from collections import defaultdict


class BondLengthQuery(QueryOperator):
    """
    Method to generate a query on bond length data.
    """

    def query(
        self,
        max_bond_length_max: Optional[float] = Query(
            None,
            description="Maximum value for the maximum bond length in the structure.",
        ),
        max_bond_length_min: Optional[float] = Query(
            None,
            description="Minimum value for the maximum bond length in the structure.",
        ),
        min_bond_length_max: Optional[float] = Query(
            None,
            description="Maximum value for the minimum bond length in the structure.",
        ),
        min_bond_length_min: Optional[float] = Query(
            None,
            description="Minimum value for the minimum bond length in the structure.",
        ),
        mean_bond_length_max: Optional[float] = Query(
            None,
            description="Maximum value for the mean bond length in the structure.",
        ),
        mean_bond_length_min: Optional[float] = Query(
            None,
            description="Minimum value for the mean bond length in the structure.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "bond_length_stats.max": [max_bond_length_min, max_bond_length_max],
            "bond_length_stats.min": [min_bond_length_min, min_bond_length_max],
            "bond_length_stats.mean": [mean_bond_length_min, mean_bond_length_max],
        }

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = [
            "bond_length_stats.max",
            "bond_length_stats.min",
            "bond_length_stats.mean",
        ]
        return [(key, False) for key in keys]


class CoordinationEnvsQuery(QueryOperator):
    """
    Method to generate a query on coordination environment data.
    """

    def query(
        self,
        coordination_envs: Optional[str] = Query(
            None,
            description="Query by coordination environments in the material composition as a comma-separated list\
 (e.g. 'Mo-S(6),S-Mo(3)')",
        ),
        coordination_envs_anonymous: Optional[str] = Query(
            None,
            description="Query by anonymous coordination environments in the material composition as a comma-separated\
 list (e.g. 'A-B(6),A-B(3)')",
        ),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if coordination_envs:
            env_list = [env.strip() for env in coordination_envs.split(",")]
            crit["coordination_envs"] = {"$all": [str(env) for env in env_list]}

        if coordination_envs_anonymous:
            env_list = [env.strip() for env in coordination_envs_anonymous.split(",")]
            crit["coordination_envs_anonymous"] = {
                "$all": [str(env) for env in env_list]
            }

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("coordination_envs", False), ("coordination_envs_anonymous", False)]



from maggma.api.resource import ReadOnlyResource
from emmet.core.grain_boundary import GrainBoundaryDoc

from emmet.api.routes.grain_boundary.query_operators import (
    GBStructureQuery,
    GBTaskIDQuery,
)
from maggma.api.query_operator import (
    PaginationQuery,
    SortQuery,
    SparseFieldsQuery,
    NumericQuery,
)
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings


def gb_resource(gb_store):
    resource = ReadOnlyResource(
        gb_store,
        GrainBoundaryDoc,
        query_operators=[
            GBTaskIDQuery(),
            NumericQuery(
                model=GrainBoundaryDoc, excluded_fields=["rotation_axis", "gb_plane"]
            ),
            GBStructureQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                GrainBoundaryDoc, default_fields=["task_id", "last_updated"]
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Grain Boundaries"],
        enable_get_by_key=False,
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT
    )

    return resource



from maggma.api.query_operator import PaginationQuery, SparseFieldsQuery
from maggma.api.resource import ReadOnlyResource

from emmet.core.phonon import PhononBSDOSDoc
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings


def phonon_bsdos_resource(phonon_bs_store):
    resource = ReadOnlyResource(
        phonon_bs_store,
        PhononBSDOSDoc,
        query_operators=[
            PaginationQuery(),
            SparseFieldsQuery(
                PhononBSDOSDoc, default_fields=["task_id", "last_updated"]
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Phonon"],
        enable_default_search=False,
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT
    )

    return resource



from maggma.api.resource import ReadOnlyResource
from emmet.core.magnetism import MagnetismDoc

from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from emmet.api.routes.magnetism.query_operators import MagneticQuery
from emmet.api.routes.materials.query_operators import MultiMaterialIDQuery
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings


def magnetism_resource(magnetism_store):
    resource = ReadOnlyResource(
        magnetism_store,
        MagnetismDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            MagneticQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(MagnetismDoc, default_fields=["material_id", "last_updated"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Magnetism"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS

from collections import defaultdict


class PossibleOxiStateQuery(QueryOperator):
    """
    Method to generate a query for possible oxidation state data.
    """

    def query(
        self,
        possible_species: str = Query(
            None,
            description="Comma delimited list of element symbols appended with oxidation states. \
                (e.g. Cr2+,O2-)",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if possible_species:
            crit["possible_species"] = {
                "$all": [specie.strip() for specie in possible_species.split(",")]
            }

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = ["possible_species"]
        return [(key, False) for key in keys]



from maggma.api.resource import ReadOnlyResource

from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery

from emmet.core.alloys import AlloyPairDoc

from emmet.api.core.settings import MAPISettings
from emmet.api.routes.alloys.query_operators import (
    MaterialIDsSearchQuery,
    FormulaSearchQuery,
)


def alloy_pairs_resource(alloy_pairs_store):
    resource = ReadOnlyResource(
        alloy_pairs_store,
        AlloyPairDoc,
        query_operators=[
            MaterialIDsSearchQuery(),
            FormulaSearchQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(AlloyPairDoc, default_fields=["pair_id"],),
        ],
        tags=["Alloys"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



import re


def make_ellipsis(text, limit=100, remove_trailing=True):
    if len(text) > limit:
        if remove_trailing:
            text = text[:limit]
            # Make results nicer by cutting incomplete words of max 15 characters.
            # For example: "This is an incomplete sente" becomes "This is an incomplete "
            trailing = re.search(r"(?<![\w\d])[\w\d]{,15}$", text)
            if trailing is not None:
                text = text[: trailing.start()] + "..."
        else:
            text = text[len(text) - limit:]
            heading = re.search(r"^[\w\d]{,15}(?![\w\d])", text)
            if heading is not None:
                text = "..." + text[heading.end():]
    return text


def mask_paragraphs(doc, limit=100):
    if isinstance(doc.get("paragraph_string", None), str):
        doc["paragraph_string"] = make_ellipsis(doc["paragraph_string"], limit=limit)

    return doc


def mask_highlights(doc, limit=100):
    if isinstance(doc.get("highlights", None), list):
        total_chars = 0
        show_hl = []

        for h_obj in doc["highlights"]:
            if total_chars >= limit:
                break
            hls = h_obj["texts"]

            # Identify where the highlighting starts.
            for i, hl in enumerate(hls):
                if hl["type"] == "hit":
                    if i > 0:
                        hls[i - 1]["value"] = make_ellipsis(
                            hls[i - 1]["value"], limit=20, remove_trailing=False
                        )
                        hls = hls[i - 1:]
                    break

            # Remove excessive chars after the hit.
            for i, hl in enumerate(hls):
                total_chars += len(hl["value"])
                if total_chars >= limit:
                    hl_limit = max(1, limit - (total_chars - len(hl["value"])))
                    hl["value"] = make_ellipsis(hl["value"], limit=hl_limit)
                    hls = hls[: i + 1]
                    break
            h_obj["texts"] = hls
            show_hl.append(h_obj)

        doc["highlights"] = show_hl

    return doc



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS


class ReconstructedQuery(QueryOperator):
    """
    Method to generate a query on whether the entry
    contains a reconstructed surface.
    """

    def query(
        self,
        has_reconstructed: Optional[bool] = Query(
            None, description="Whether the entry has a reconstructed surface.",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if has_reconstructed is not None:
            crit.update({"has_reconstructed": has_reconstructed})

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("has_reconstructed", False)]



from typing import Dict, List
from maggma.api.resource.core import Resource
from pymatgen.core import __version__ as pmg_version  # type: ignore
from emmet.api.core.settings import MAPISettings
from emmet.api import __version__ as api_version
from fastapi.openapi.utils import get_openapi
from maggma.api.API import API


class MAPI(API):
    """
    Core Materials API that orchestrates resources together
    """

    def __init__(
        self,
        resources: Dict[str, List[Resource]],
        title="Materials Project API",
        version=api_version,
        debug=False,
        heartbeat_meta={
            "pymatgen": pmg_version,
            "db_version": MAPISettings().DB_VERSION,
            "suffix": MAPISettings().DB_NAME_SUFFIX,
        },
        description=None,
        tags_meta=None,
    ):
        super().__init__(
            resources=resources,
            title=title,
            version=version,
            debug=debug,
            heartbeat_meta=heartbeat_meta,
            description=description,
            tags_meta=tags_meta,
        )

    @property
    def app(self):
        """
        App server for the cluster manager
        """
        app = super().app

        def custom_openapi():
            openapi_schema = get_openapi(
                title=self.title,
                version=self.version,
                routes=app.routes,
                description=self.description,
                tags=self.tags_meta,
            )

            openapi_schema["components"]["securitySchemes"] = {
                "ApiKeyAuth": {
                    "descriptions": "MP API key to authorize requests",
                    "name": "X-API-KEY",
                    "in": "header",
                    "type": "apiKey",
                }
            }

            openapi_schema["security"] = [{"ApiKeyAuth": []}]

            openapi_schema["info"]["x-logo"] = {
                "url": "https://raw.githubusercontent.com/materialsproject/emmet/master/emmet-api/emmet/api/core/assets/mp_logo_small.png"  # noqa: E501
            }

            app.openapi_schema = openapi_schema
            return app.openapi_schema

        app.openapi = custom_openapi

        return app



from maggma.api.resource import ReadOnlyResource
from emmet.core.dois import DOIDoc
from emmet.api.core.global_header import GlobalHeaderProcessor
from maggma.api.query_operator import PaginationQuery, SparseFieldsQuery
from emmet.api.core.settings import MAPISettings


def dois_resource(dois_store):
    resource = ReadOnlyResource(
        dois_store,
        DOIDoc,
        query_operators=[
            PaginationQuery(),
            SparseFieldsQuery(DOIDoc, default_fields=["task_id", "doi"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["DOIs"],
        enable_default_search=False,
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from maggma.api.resource import ReadOnlyResource
from emmet.core.molecules_jcesr import MoleculesDoc

from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from emmet.api.routes.molecules.query_operators import (
    MoleculeBaseQuery,
    MoleculeElementsQuery,
    MoleculeFormulaQuery,
)
from emmet.api.routes.tasks.query_operators import MultipleTaskIDsQuery
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings


def molecules_resource(molecules_store):
    resource = ReadOnlyResource(
        molecules_store,
        MoleculesDoc,
        query_operators=[
            MoleculeBaseQuery(),
            MoleculeElementsQuery(),
            MoleculeFormulaQuery(),
            MultipleTaskIDsQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(MoleculesDoc, default_fields=["task_id"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Molecules"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT
    )

    return resource



from maggma.api.query_operator.dynamic import NumericQuery
from maggma.api.resource import ReadOnlyResource
from emmet.core.electronic_structure import ElectronicStructureDoc
from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery

from emmet.api.routes.materials.query_operators import (
    ElementsQuery,
    FormulaQuery,
    ChemsysQuery,
    DeprecationQuery,
    MultiMaterialIDQuery
)

from emmet.api.routes.electronic_structure.query_operators import (
    ESSummaryDataQuery,
    BSDataQuery,
    DOSDataQuery,
    ObjectQuery,
)
from emmet.core.electronic_structure import BSObjectDoc, DOSObjectDoc
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings

timeout = MAPISettings().TIMEOUT


def es_resource(es_store):
    resource = ReadOnlyResource(
        es_store,
        ElectronicStructureDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            ESSummaryDataQuery(),
            FormulaQuery(),
            ChemsysQuery(),
            ElementsQuery(),
            NumericQuery(model=ElectronicStructureDoc),
            DeprecationQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                ElectronicStructureDoc, default_fields=["material_id", "last_updated"]
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Electronic Structure"],
        disable_validation=True,
        timeout=timeout,
    )

    return resource


def bs_resource(es_store):
    resource = ReadOnlyResource(
        es_store,
        ElectronicStructureDoc,
        query_operators=[
            BSDataQuery(),
            DeprecationQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                ElectronicStructureDoc,
                default_fields=["material_id", "last_updated", "bandstructure"],
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Electronic Structure"],
        enable_get_by_key=False,
        sub_path="/bandstructure/",
        disable_validation=True,
        timeout=timeout,
    )

    return resource


def bs_obj_resource(s3_store):
    resource = ReadOnlyResource(
        s3_store,
        BSObjectDoc,
        query_operators=[
            ObjectQuery(),
            SparseFieldsQuery(BSObjectDoc, default_fields=["task_id", "last_updated"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Electronic Structure"],
        enable_get_by_key=False,
        enable_default_search=True,
        sub_path="/bandstructure/object/",
        disable_validation=True,
    )
    return resource


def dos_resource(es_store):
    resource = ReadOnlyResource(
        es_store,
        ElectronicStructureDoc,
        query_operators=[
            DOSDataQuery(),
            DeprecationQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                ElectronicStructureDoc,
                default_fields=["material_id", "last_updated", "dos"],
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Electronic Structure"],
        enable_get_by_key=False,
        sub_path="/dos/",
        disable_validation=True,
        timeout=timeout
    )

    return resource


def dos_obj_resource(s3_store):
    resource = ReadOnlyResource(
        s3_store,
        DOSObjectDoc,
        query_operators=[
            ObjectQuery(),
            SparseFieldsQuery(DOSObjectDoc, default_fields=["task_id", "last_updated"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Electronic Structure"],
        enable_get_by_key=False,
        enable_default_search=True,
        sub_path="/dos/object/",
        disable_validation=True,
    )
    return resource



from collections import defaultdict
from typing import List

from fastapi import HTTPException
from pymatgen.analysis.structure_analyzer import oxide_type
from pymatgen.core import Structure
from pymatgen.core.trajectory import Trajectory
from pymatgen.entries.computed_entries import ComputedStructureEntry


def calcs_reversed_to_trajectory(calcs_reversed: List[dict]):
    """
    Converts data from calc_reversed to pymatgen Trajectory objects
    that contain structure, energy, force and stress data for each
    ionic step.

    Args:
        calcs_reversed: List of dictionaries in calcs_reversed entry
            of a task document.
    """
    trajectories = []

    for calculation in calcs_reversed:
        structures = []
        frame_props = defaultdict(list)  # type: dict

        steps = calculation.get("output", {}).get("ionic_steps", None)

        if steps is None:
            raise HTTPException(
                status_code=404, detail="No ionic step data found for task"
            )
        else:
            for step in steps:

                structure_dict = step.get("structure", None)

                if structure_dict is not None:
                    structure = Structure.from_dict(structure_dict)

                structures.append(structure)

                frame_props["e_fr_energy"].append(step.get("e_fr_energy", None))
                frame_props["e_wo_entrp"].append(step.get("e_wo_entrp", None))
                frame_props["e_0_energy"].append(step.get("e_0_energy", None))
                frame_props["forces"].append(step.get("forces", None))
                frame_props["stresses"].append(step.get("stress", None))
                frame_props["electronic_steps"].append(
                    step.get("electronic_steps", None)
                )

            traj = Trajectory.from_structures(
                structures, frame_properties=frame_props, time_step=None
            ).as_dict()
            trajectories.append(traj)

    return trajectories


def task_to_entry(doc: dict, include_structure: bool = True):
    """Turns a Task Doc into a ComputedStructureEntry"""

    input = doc.get("input", None)
    output = doc.get("output", None)

    try:

        output_struct = Structure.from_dict(output["structure"])

        entry_dict = {
            "correction": 0.0,
            "entry_id": doc["task_id"],
            "composition": output_struct.composition,
            "energy": output.get("energy", None),
            "parameters": {
                "potcar_spec": input.get("potcar_spec", None),
                "is_hubbard": input.get("is_hubbard", None),
                "hubbards": input.get("hubbards", None),
                "run_type": str(doc["run_type"]) if "run_type" in doc else None,
                "task_type": str(doc["task_type"]) if "task_type" in doc else None,
            },
            "data": {
                "oxide_type": oxide_type(output_struct),
                "aspherical": input.get("parameters", {}).get("LASPH", False),
                "last_updated": doc["last_updated"] if "last_updated" in doc else None,
                "completed_at": doc["completed_at"] if "completed_at" in doc else None,
            },
            "structure": output_struct,
        }

        return ComputedStructureEntry.from_dict(entry_dict).as_dict()

    except (AttributeError, KeyError, TypeError):
        return "Problem obtaining entry for {}. It might be missing necessary output structure information.".format(
            doc["task_id"]
        )



from maggma.api.resource import ReadOnlyResource
from emmet.core.xas import XASDoc

from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from emmet.api.routes.materials.query_operators import (
    ElementsQuery,
    FormulaQuery,
    ChemsysQuery,
)
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.routes.xas.query_operators import XASQuery, XASTaskIDQuery
from emmet.api.core.settings import MAPISettings


def xas_resource(xas_store):
    resource = ReadOnlyResource(
        xas_store,
        XASDoc,
        query_operators=[
            FormulaQuery(),
            ChemsysQuery(),
            ElementsQuery(),
            XASQuery(),
            XASTaskIDQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                XASDoc,
                default_fields=[
                    "xas_id",
                    "task_id",
                    "edge",
                    "absorbing_element",
                    "formula_pretty",
                    "spectrum_type",
                    "last_updated",
                ],
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["XAS"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from maggma.api.resource import ReadOnlyResource
from emmet.core.chemenv import ChemEnvDoc
from emmet.api.core.global_header import GlobalHeaderProcessor
from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from emmet.api.routes.materials.query_operators import MultiMaterialIDQuery
from emmet.api.core.settings import MAPISettings


def chemenv_resource(chemenv_store):
    resource = ReadOnlyResource(
        chemenv_store,
        ChemEnvDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                ChemEnvDoc,
                default_fields=["material_id", "formula_pretty", "last_updated"],
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Chemical Environment"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT
    )

    return resource



from maggma.api.resource import ReadOnlyResource
from maggma.api.query_operator import PaginationQuery, SparseFieldsQuery
from emmet.core.charge_density import ChgcarDataDoc
from emmet.api.routes.charge_density.query_operators import ChgcarTaskIDQuery
from emmet.api.core.global_header import GlobalHeaderProcessor


def charge_density_resource(s3_store):
    resource = ReadOnlyResource(
        s3_store,
        ChgcarDataDoc,
        query_operators=[
            ChgcarTaskIDQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                ChgcarDataDoc, default_fields=["task_id", "last_updated", "fs_id"]
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Charge Density"],
        enable_default_search=True,
        enable_get_by_key=False,
        disable_validation=True,
    )

    return resource


def charge_density_url_resource(s3_store):
    resource = ReadOnlyResource(
        s3_store,
        ChgcarDataDoc,
        key_fields=["task_id", "fs_id", "url", "s3_url_prefix", "requested_datetime", "expiry_datetime"],
        tags=["Charge Density"],
        enable_default_search=False,
        enable_get_by_key=True,
        disable_validation=True,
    )

    return resource



from maggma.api.resource.read_resource import ReadOnlyResource
from maggma.api.resource.post_resource import PostOnlyResource
from maggma.api.resource.aggregation import AggregationResource


from emmet.core.vasp.material import MaterialsDoc
from emmet.core.find_structure import FindStructure
from emmet.core.formula_autocomplete import FormulaAutocomplete

from maggma.api.query_operator import (
    PaginationQuery,
    SparseFieldsQuery,
    SortQuery,
    NumericQuery,
)

from emmet.api.routes.materials.hint_scheme import MaterialsHintScheme
from emmet.api.routes.materials.query_operators import (
    ElementsQuery,
    FormulaQuery,
    ChemsysQuery,
    DeprecationQuery,
    SymmetryQuery,
    MultiTaskIDQuery,
    FindStructureQuery,
    FormulaAutoCompleteQuery,
    MultiMaterialIDQuery,
)
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings

timeout = MAPISettings().TIMEOUT


def find_structure_resource(materials_store):
    resource = PostOnlyResource(
        materials_store,
        FindStructure,
        key_fields=["structure", "task_id"],
        query_operators=[FindStructureQuery()],
        tags=["Materials"],
        sub_path="/find_structure/",
        timeout=timeout,
    )

    return resource


def formula_autocomplete_resource(formula_autocomplete_store):
    resource = AggregationResource(
        formula_autocomplete_store,
        FormulaAutocomplete,
        pipeline_query_operator=FormulaAutoCompleteQuery(),
        tags=["Materials"],
        sub_path="/formula_autocomplete/",
        header_processor=GlobalHeaderProcessor(),
        timeout=timeout,
    )

    return resource


def materials_resource(materials_store):

    resource = ReadOnlyResource(
        materials_store,
        MaterialsDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            FormulaQuery(),
            ChemsysQuery(),
            ElementsQuery(),
            MultiTaskIDQuery(),
            SymmetryQuery(),
            DeprecationQuery(),
            NumericQuery(model=MaterialsDoc),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(MaterialsDoc, default_fields=["material_id", "formula_pretty", "last_updated"],),
        ],
        header_processor=GlobalHeaderProcessor(),
        hint_scheme=MaterialsHintScheme(),
        tags=["Materials"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from maggma.api.resource import ReadOnlyResource
from maggma.api.query_operator import PaginationQuery, SparseFieldsQuery
from emmet.api.routes.materials.query_operators import DeprecationQuery, MultiMaterialIDQuery
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.core.provenance import ProvenanceDoc
from emmet.api.core.settings import MAPISettings


def provenance_resource(provenance_store):
    resource = ReadOnlyResource(
        provenance_store,
        ProvenanceDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            DeprecationQuery(),
            PaginationQuery(),
            SparseFieldsQuery(ProvenanceDoc, default_fields=["material_id", "last_updated"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Provenance"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from maggma.api.resource import SubmissionResource
from maggma.api.query_operator import PaginationQuery
from emmet.core.mpcomplete import MPCompleteDoc, MPCompleteDataStatus
from emmet.api.routes.mpcomplete.query_operator import (
    MPCompletePostQuery,
    MPCompleteGetQuery,
)
from emmet.api.core.settings import MAPISettings


def mpcomplete_resource(mpcomplete_store):
    resource = SubmissionResource(
        mpcomplete_store,
        MPCompleteDoc,
        post_query_operators=[MPCompletePostQuery()],
        get_query_operators=[MPCompleteGetQuery(), PaginationQuery()],
        tags=["MPComplete"],
        state_enum=MPCompleteDataStatus,
        default_state=MPCompleteDataStatus.submitted.value,
        calculate_submission_id=True,
        include_in_schema=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from itertools import permutations
from typing import Optional

from emmet.core.symmetry import CrystalSystem
from fastapi import Body, HTTPException, Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS
from emmet.api.routes.materials.utils import formula_to_criteria, chemsys_to_criteria
from pymatgen.analysis.structure_matcher import ElementComparator, StructureMatcher
from pymatgen.core.composition import Composition, CompositionError
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Structure


class FormulaQuery(QueryOperator):
    """
    Factory method to generate a dependency for querying by
        formula or chemical system with wild cards.
    """

    def query(
        self,
        formula: Optional[str] = Query(
            None,
            description="Query by formula including anonymized formula or by including wild cards. \
A comma delimited string list of anonymous formulas or regular formulas can also be provided.",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if formula:
            crit.update(formula_to_criteria(formula))

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = ["formula_pretty", "formula_anonymous", "composition_reduced"]
        return [(key, False) for key in keys]


class ChemsysQuery(QueryOperator):
    """
    Factory method to generate a dependency for querying by
        chemical system with wild cards.
    """

    def query(
        self,
        chemsys: Optional[str] = Query(
            None,
            description="A comma delimited string list of chemical systems. \
Wildcards for unknown elements only supported for single chemsys queries",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if chemsys:
            crit.update(chemsys_to_criteria(chemsys))

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = ["chemsys", "elements", "nelements"]
        return [(key, False) for key in keys]


class ElementsQuery(QueryOperator):
    """
    Factory method to generate a dependency for querying by element data
    """

    def query(
        self,
        elements: Optional[str] = Query(
            None, description="Query by elements in the material composition as a comma-separated list",
        ),
        exclude_elements: Optional[str] = Query(
            None, description="Query by excluded elements in the material composition as a comma-separated list",
        ),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if elements or exclude_elements:
            crit["elements"] = {}

        if elements:
            try:
                element_list = [Element(e) for e in elements.strip().split(",")]
            except ValueError:
                raise HTTPException(
                    status_code=400, detail="Please provide a comma-seperated list of elements",
                )

            crit["elements"]["$all"] = [str(el) for el in element_list]

        if exclude_elements:
            try:
                element_list = [Element(e) for e in exclude_elements.strip().split(",")]
            except ValueError:
                raise HTTPException(
                    status_code=400, detail="Please provide a comma-seperated list of elements",
                )
            crit["elements"]["$nin"] = [str(el) for el in element_list]

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("elements", False)]


class DeprecationQuery(QueryOperator):
    """
    Method to generate a deprecation state query
    """

    def query(
        self, deprecated: Optional[bool] = Query(False, description="Whether the material is marked as deprecated",),
    ) -> STORE_PARAMS:

        crit = {}

        if deprecated is not None:
            crit.update({"deprecated": deprecated})

        return {"criteria": crit}


class SymmetryQuery(QueryOperator):
    """
    Method to generate a query on symmetry information
    """

    def query(
        self,
        crystal_system: Optional[CrystalSystem] = Query(None, description="Crystal system of the material",),
        spacegroup_number: Optional[int] = Query(None, description="Space group number of the material",),
        spacegroup_symbol: Optional[str] = Query(None, description="Space group symbol of the material",),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if crystal_system:
            crit.update({"symmetry.crystal_system": str(crystal_system.value)})

        if spacegroup_number:
            crit.update({"symmetry.number": spacegroup_number})

        if spacegroup_symbol:
            crit.update({"symmetry.symbol": spacegroup_symbol})

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = ["symmetry.crystal_system", "symmetry.number", "symmetry.symbol"]
        return [(key, False) for key in keys]


class MultiTaskIDQuery(QueryOperator):
    """
    Method to generate a query for different task_ids
    """

    def query(
        self, task_ids: Optional[str] = Query(None, description="Comma-separated list of task_ids to query on"),
    ) -> STORE_PARAMS:

        crit = {}

        if task_ids:
            crit.update({"task_ids": {"$in": [task_id.strip() for task_id in task_ids.split(",")]}})

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("task_ids", False)]


class MultiMaterialIDQuery(QueryOperator):
    """
    Method to generate a query for different root-level material_id values
    """

    def query(
        self,
        material_ids: Optional[str] = Query(None, description="Comma-separated list of material_id values to query on"),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if material_ids:

            mpids_list = [material_id.strip() for material_id in material_ids.split(",")]

            if len(mpids_list) == 1:
                crit.update({"material_id": mpids_list[0]})
            else:
                crit.update({"material_id": {"$in": mpids_list}})

        return {"criteria": crit}


class FindStructureQuery(QueryOperator):
    """
    Method to generate a find structure query
    """

    def query(
        self,
        structure: Structure = Body(..., description="Pymatgen structure object to query with",),
        ltol: float = Query(0.2, description="Fractional length tolerance. Default is 0.2.",),
        stol: float = Query(
            0.3,
            description="Site tolerance. Defined as the fraction of the average free \
                    length per atom := ( V / Nsites ) ** (1/3). Default is 0.3.",
        ),
        angle_tol: float = Query(5, description="Angle tolerance in degrees. Default is 5 degrees.",),
        _limit: int = Query(
            1, description="Maximum number of matches to show. Defaults to 1, only showing the best match.",
        ),
    ) -> STORE_PARAMS:

        self.ltol = ltol
        self.stol = stol
        self.angle_tol = angle_tol
        self._limit = _limit
        self.structure = structure

        crit = {}

        try:
            s = Structure.from_dict(structure)
        except Exception:
            raise HTTPException(
                status_code=404, detail="Body cannot be converted to a pymatgen structure object.",
            )

        crit.update({"composition_reduced": dict(s.composition.to_reduced_dict)})

        return {"criteria": crit}

    def post_process(self, docs, query):

        s1 = Structure.from_dict(self.structure)

        m = StructureMatcher(
            ltol=self.ltol,
            stol=self.stol,
            angle_tol=self.angle_tol,
            primitive_cell=True,
            scale=True,
            attempt_supercell=False,
            comparator=ElementComparator(),
        )

        matches = []

        for doc in docs:

            s2 = Structure.from_dict(doc["structure"])
            matched = m.fit(s1, s2)

            if matched:
                rms = m.get_rms_dist(s1, s2)

                matches.append(
                    {
                        "material_id": doc["material_id"],
                        "normalized_rms_displacement": rms[0],
                        "max_distance_paired_sites": rms[1],
                    }
                )

        response = sorted(
            matches[: self._limit], key=lambda x: (x["normalized_rms_displacement"], x["max_distance_paired_sites"],),
        )

        return response

    def ensure_indexes(self):  # pragma: no cover
        return [("composition_reduced", False)]


class FormulaAutoCompleteQuery(QueryOperator):
    """
    Method to generate a formula autocomplete query
    """

    def query(
        self,
        formula: str = Query(..., description="Human readable chemical formula.",),
        limit: int = Query(10, description="Maximum number of matches to show. Defaults to 10.",),
    ) -> STORE_PARAMS:

        self.formula = formula
        self.limit = limit

        try:
            comp = Composition(formula)
        except (CompositionError, ValueError):
            raise HTTPException(
                status_code=400, detail="Invalid formula provided.",
            )

        ind_str = []
        eles = []

        if len(comp) == 1:
            d = comp.get_integer_formula_and_factor()

            s = d[0] + str(int(d[1])) if d[1] != 1 else d[0]

            ind_str.append(s)
            eles.append(d[0])
        else:

            comp_red = comp.reduced_composition.items()

            for (i, j) in comp_red:

                if j != 1:
                    ind_str.append(i.name + str(int(j)))
                else:
                    ind_str.append(i.name)

                eles.append(i.name)

        final_terms = ["".join(entry) for entry in permutations(ind_str)]

        pipeline = [
            {"$search": {"index": "formula_autocomplete", "text": {"path": "formula_pretty", "query": final_terms}}},
            {"$project": {"_id": 0, "formula_pretty": 1, "elements": 1, "length": {"$strLenCP": "$formula_pretty"}}},
            {"$match": {"length": {"$gte": len(final_terms[0])}, "elements": {"$all": eles}}},
            {"$limit": limit},
            {"$sort": {"length": 1}},
            {"$project": {"elements": 0, "length": 0}},
        ]

        return {"pipeline": pipeline}

    def ensure_indexes(self):  # pragma: no cover
        return [("formula_pretty", False)]



from maggma.api.resource import ReadOnlyResource
from emmet.core.elasticity_legacy import ElasticityDoc
from emmet.api.core.global_header import GlobalHeaderProcessor
from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from emmet.api.routes.elasticity.query_operators import (
    ElasticityChemsysQuery,
    BulkModulusQuery,
    ShearModulusQuery,
    PoissonQuery,
)

from emmet.api.core.settings import MAPISettings


def elasticity_resource(elasticity_store):
    resource = ReadOnlyResource(
        elasticity_store,
        ElasticityDoc,
        query_operators=[
            ElasticityChemsysQuery(),
            BulkModulusQuery(),
            ShearModulusQuery(),
            PoissonQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                ElasticityDoc, default_fields=["task_id", "pretty_formula"],
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Elasticity"],
        disable_validation=False,
        timeout=MAPISettings().TIMEOUT
    )

    return resource



"""
This script converts synthesis recipes data fetched directly
from the public repo of synthesis recipes
(https://github.com/CederGroupHub/text-mined-synthesis_public)
into MP compatible formats.
"""
import json
import sys
import logging

from pymatgen.core import Composition
from pymatgen.core.composition import CompositionError


logger = logging.getLogger(__name__)


def string2comp(x):
    """Convert string material formulas into pymatgen Compositions."""
    # TODO: if a material contains multiple parts, this function
    #  only takes the first part. This is not the optimal solution,
    #  and should be resolved in the future.
    formula = x.split("·")

    # return reduce(add, [Composition(x) for x in formula])
    return Composition(formula[0])


def convert_recipe(recipe):
    """Convert an entire synthesis recipe."""
    targets_string = recipe["targets_string"]
    try:
        target_comps = [string2comp(x) for x in targets_string]
    except (CompositionError, ValueError):
        logger.error(f"Cannot process materials: {targets_string}")
        raise

    recipe["targets_formula"] = [x.formula for x in target_comps]
    recipe["targets_formula_s"] = [x.reduced_formula for x in target_comps]
    del recipe["targets_string"]

    recipe["precursors_formula"] = []
    recipe["precursors_formula_s"] = []
    for precursor in recipe["precursors"]:
        try:
            comp = string2comp(precursor["material_formula"])
        except (CompositionError, ValueError):
            logger.info("Cannot process precursor material: " + precursor["material_formula"])
            continue
        recipe["precursors_formula"].append(comp.formula)
        recipe["precursors_formula_s"].append(comp.reduced_formula)

    return recipe


def convert_json_public_repo(src_json, dst_json):
    """
    Convert the public synthesis recipes dataset (in a json file)
    into a format as json file which can be imported into the MP database.
    """
    with open(src_json) as f:
        data = json.load(f)
        recipes = data["reactions"]
        logger.info("Loaded %s recipes, version %s" % (len(recipes), data["release_date"]))

    converted = []
    for recipe in recipes:
        try:
            convert_recipe(recipe)
            converted.append(recipe)
        except (CompositionError, ValueError, IndexError):
            pass

    logger.info("Converted %d recipes" % (len(converted),))
    with open(dst_json, "w") as f:
        json.dump(converted, f)


if __name__ == "__main__":
    convert_json_public_repo(sys.argv[1], sys.argv[2])



from typing import Literal
from pydantic import BaseSettings, Field
from emmet.api import __file__ as root_dir
import os


class MAPISettings(BaseSettings):
    """
    Special class to store settings for MAPI
    python module
    """

    DEBUG: bool = Field(False, description="Turns on debug mode for MAPI")

    TEST_FILES: str = Field(
        os.path.join(
            os.path.dirname(os.path.abspath(root_dir)), "../../../tests/test_files"
        ),
        description="Directory with test files",
    )

    DB_VERSION: str = Field("2021.11.10", description="Database version")

    DB_NAME_SUFFIX: Literal["blue", "green"] = Field(None, description="Database name suffix. Either blue or green.")

    TIMEOUT: int = Field(
        20, description="Number of seconds to wait for pymongo operations before raising a timeout error."
    )

    class Config:
        env_prefix = "MAPI_"



from maggma.api.resource import ReadOnlyResource
from emmet.core.bonds import BondingDoc

from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery

from emmet.api.routes.bonds.query_operators import (
    BondLengthQuery,
    CoordinationEnvsQuery,
)
from emmet.api.routes.materials.query_operators import MultiMaterialIDQuery
from emmet.api.core.settings import MAPISettings
from emmet.api.core.global_header import GlobalHeaderProcessor


def bonds_resource(bonds_store):
    resource = ReadOnlyResource(
        bonds_store,
        BondingDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            BondLengthQuery(),
            CoordinationEnvsQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(BondingDoc, default_fields=["material_id", "last_updated"],),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Bonds"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS

from collections import defaultdict


class SubstrateStructureQuery(QueryOperator):
    """
    Method to generate a query for film and substrate data.
    """

    def query(
        self,
        film_orientation: Optional[str] = Query(
            None,
            description="Comma separated integers defining the film surface orientation.",
        ),
        substrate_orientation: Optional[str] = Query(
            None,
            description="Comma separated integers defining the substrate surface orientation.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if film_orientation:
            crit["film_orient"] = " ".join(
                [entry.strip() for entry in film_orientation.split(",")]
            )

        if substrate_orientation:
            crit["orient"] = " ".join(
                [entry.strip() for entry in substrate_orientation.split(",")]
            )

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = ["film_id", "sub_id", "sub_form", "film_orient", "orient"]
        return [(key, False) for key in keys]


class EnergyAreaQuery(QueryOperator):
    """
    Method to generate a query for ranges of substrate
    elastic energies and minimum coincident areas.
    """

    def query(
        self,
        area_max: Optional[float] = Query(
            None,
            description="Maximum value for the minimum coincident interface area in Å².",
        ),
        area_min: Optional[float] = Query(
            None,
            description="Minimum value for the minimum coincident interface area in Å².",
        ),
        energy_max: Optional[float] = Query(
            None, description="Maximum value for the energy in meV.",
        ),
        energy_min: Optional[float] = Query(
            None, description="Minimum value for the energy in meV.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "area": [area_min, area_max],
            "energy": [energy_min, energy_max],
        }

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = ["area", "energy"]
        return [(key, False) for key in keys]



from pymatgen.core import Composition
from pymatgen.core.periodic_table import DummySpecies
from typing import Dict
from fastapi import HTTPException


def electrodes_formula_to_criteria(formulas: str) -> Dict:
    """
    Santizes formula into a dictionary to search with wild cards
    over electrodes data

    Arguments:
        formula: formula with wildcards in it for unknown elements

    Returns:
        Mongo style search criteria for this formula
    """
    dummies = "ADEGJLMQRXZ"

    formula_list = [formula.strip() for formula in formulas.split(",")]

    if "*" in formulas:
        if len(formula_list) > 1:
            raise HTTPException(
                status_code=400,
                detail="Wild cards only supported for single formula queries.",
            )
        else:
            # Wild card in formula
            nstars = formulas.count("*")

            formula_dummies = formulas.replace("*", "{}").format(*dummies[:nstars])

            try:
                integer_formula = Composition(
                    formula_dummies
                ).get_integer_formula_and_factor()[0]
            except (ValueError, IndexError):
                raise HTTPException(
                    status_code=400,
                    detail="Problem processing one or more provided formulas.",
                )

            comp = Composition(integer_formula).reduced_composition
            crit = dict()  # type: dict
            crit[
                "entries_composition_summary.all_formula_anonymous"
            ] = comp.anonymized_formula

            real_elts = [
                str(e)
                for e in comp.elements
                if not e.as_dict().get("element", "A") in dummies
            ]

            for el, n in comp.to_reduced_dict.items():
                if el in real_elts:
                    crit[
                        f"entries_composition_summary.all_composition_reduced.{el}"
                    ] = n

            return crit

    else:
        try:
            if any(
                isinstance(el, DummySpecies)
                for formula in formula_list
                for el in Composition(formula)
            ):
                # Assume fully anonymized formula
                if len(formula_list) == 1:
                    return {
                        "entries_composition_summary.all_formula_anonymous": Composition(
                            formula_list[0]
                        ).anonymized_formula
                    }
                else:
                    return {
                        "entries_composition_summary.all_formula_anonymous": {
                            "$in": [
                                Composition(formula).anonymized_formula
                                for formula in formula_list
                            ]
                        }
                    }

            else:
                if len(formula_list) == 1:
                    comp = Composition(formula_list[0])
                    nele = len(comp)
                    # Paranoia below about floating-point "equality"
                    crit = {}
                    crit["nelements"] = {"$in": [nele, nele - 1]}  # type: ignore

                    for el, n in comp.to_reduced_dict.items():
                        crit[
                            f"entries_composition_summary.all_composition_reduced.{el}"
                        ] = n

                    return crit
                else:
                    return {
                        "entries_composition_summary.all_formulas": {
                            "$in": [
                                Composition(formula).reduced_formula
                                for formula in formula_list
                            ]
                        }
                    }
        except (ValueError, IndexError):
            raise HTTPException(
                status_code=400,
                detail="Problem processing one or more provided formulas.",
            )


def electrodes_chemsys_to_criteria(chemsys: str) -> Dict:
    """
    Santizes chemsys into a dictionary to search with wild cards
    over electrodes data

    Arguments:
        chemsys: A comma delimited string list ofchemical systems
            with wildcards in it for unknown elements

    Returns:
        Mongo style search criteria for this formula
    """

    crit = {}  # type: dict

    chemsys_list = [chemsys_val.strip() for chemsys_val in chemsys.split(",")]

    if "*" in chemsys:
        if len(chemsys_list) > 1:
            raise HTTPException(
                status_code=400,
                detail="Wild cards only supported for single chemsys queries.",
            )
        else:
            eles = chemsys_list[0].split("-")
            neles = len(eles)

            crit["nelements"] = {"$in": [neles, neles - 1]}
            crit["entries_composition_summary.all_elements"] = {
                "$all": [ele for ele in eles if ele != "*"]
            }

            if crit["entries_composition_summary.all_elements"]["$all"] == []:
                del crit["entries_composition_summary.all_elements"]

            return crit
    else:
        query_vals = []
        for chemsys_val in chemsys_list:
            eles = chemsys_val.split("-")
            sorted_chemsys = "-".join(sorted(eles))
            query_vals.append(sorted_chemsys)

        if len(query_vals) == 1:
            crit["entries_composition_summary.all_chemsys"] = query_vals[0]
        else:
            crit["entries_composition_summary.all_chemsys"] = {"$in": query_vals}

        return crit



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator


class IsStableQuery(QueryOperator):
    """
    Method to generate a query on whether a material is stable
    """

    def query(
        self,
        is_stable: Optional[bool] = Query(
            None, description="Whether the material is stable."
        ),
    ):

        crit = {}

        if is_stable is not None:
            crit["is_stable"] = is_stable

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = self._keys_from_query()
        return [(key, False) for key in keys]



from pymatgen.core import Composition
from pymatgen.core.periodic_table import DummySpecies
from typing import Dict
from fastapi import HTTPException


def formula_to_criteria(formulas: str) -> Dict:
    """
    Santizes formula into a dictionary to search with wild cards

    Arguments:
        formula: formula with wildcards in it for unknown elements

    Returns:
        Mongo style search criteria for this formula
    """
    dummies = "ADEGJLMQRXZ"

    formula_list = [formula.strip() for formula in formulas.split(",")]

    if "*" in formulas:
        if len(formula_list) > 1:
            raise HTTPException(
                status_code=400,
                detail="Wild cards only supported for single formula queries.",
            )
        else:
            # Wild card in formula
            nstars = formulas.count("*")

            formula_dummies = formulas.replace("*", "{}").format(*dummies[:nstars])

            try:
                integer_formula = Composition(
                    formula_dummies
                ).get_integer_formula_and_factor()[0]
            except ValueError:
                raise HTTPException(
                    status_code=400,
                    detail="Problem processing formula with wild cards.",
                )

            comp = Composition(integer_formula).reduced_composition
            crit = dict()
            crit["formula_anonymous"] = comp.anonymized_formula
            real_elts = [
                str(e)
                for e in comp.elements
                if not e.as_dict().get("element", "A") in dummies
            ]

            for el, n in comp.to_reduced_dict.items():
                if el in real_elts:
                    crit[f"composition_reduced.{el}"] = n

            return crit

    else:
        try:
            composition_list = [Composition(formula) for formula in formula_list]
        except ValueError:
            raise HTTPException(
                status_code=400,
                detail="Problem processing one or more provided formulas.",
            )

        if any(
            isinstance(el, DummySpecies) for comp in composition_list for el in comp
        ):
            # Assume fully anonymized formula
            if len(formula_list) == 1:
                return {"formula_anonymous": composition_list[0].anonymized_formula}
            else:
                return {
                    "formula_anonymous": {
                        "$in": [comp.anonymized_formula for comp in composition_list]
                    }
                }

        else:

            if len(formula_list) == 1:

                comp = composition_list[0]
                # Paranoia below about floating-point "equality"
                crit = {}
                crit["nelements"] = len(comp)  # type: ignore

                try:
                    for el, n in comp.to_reduced_dict.items():
                        crit[f"composition_reduced.{el}"] = n
                except IndexError:
                    raise HTTPException(
                        status_code=400,
                        detail="Problem processing one or more provided formulas.",
                    )

                return crit
            else:
                return {
                    "formula_pretty": {
                        "$in": [comp.reduced_formula for comp in composition_list]
                    }
                }


def chemsys_to_criteria(chemsys: str) -> Dict:
    """
    Santizes chemsys into a dictionary to search with wild cards

    Arguments:
        chemsys: A comma delimited string list of chemical systems
            with wildcards in it for unknown elements

    Returns:
        Mongo style search criteria for this formula
    """

    crit = {}  # type: dict

    chemsys_list = [chemsys_val.strip() for chemsys_val in chemsys.split(",")]

    if "*" in chemsys:
        if len(chemsys_list) > 1:
            raise HTTPException(
                status_code=400,
                detail="Wild cards only supported for single chemsys queries.",
            )
        else:
            eles = chemsys_list[0].split("-")

            crit["nelements"] = len(eles)
            crit["elements"] = {"$all": [ele for ele in eles if ele != "*"]}

            if crit["elements"]["$all"] == []:
                del crit["elements"]

            return crit
    else:
        query_vals = []
        for chemsys_val in chemsys_list:
            eles = chemsys_val.split("-")
            sorted_chemsys = "-".join(sorted(eles))
            query_vals.append(sorted_chemsys)

        if len(query_vals) == 1:
            crit["chemsys"] = query_vals[0]
        else:
            crit["chemsys"] = {"$in": query_vals}

        return crit



from multiprocessing.sharedctypes import Value
from typing import Optional
from fastapi import Query, HTTPException
from pymatgen.core.periodic_table import Element
from pymatgen.core import Composition
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS

from collections import defaultdict


class MoleculeElementsQuery(QueryOperator):
    """
    Method to generate a query on molecules using a list of elements
    """

    def query(
        self,
        elements: Optional[str] = Query(
            None,
            description="Query by elements in the material composition as a comma-separated list",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if elements:
            try:
                element_list = [Element(e.strip()) for e in elements.split(",")]
            except ValueError:
                raise HTTPException(
                    status_code=400,
                    detail="Problem processing one or more provided elements.",
                )
            crit["elements"] = {"$all": [str(el) for el in element_list]}

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("elements", False)]


class MoleculeBaseQuery(QueryOperator):
    """
    Method to generate a query over molecule data.
    """

    def query(
        self,
        nelements_max: Optional[float] = Query(
            None, description="Maximum value for the number of elements.",
        ),
        nelements_min: Optional[float] = Query(
            None, description="Minimum value for the number of elements.",
        ),
        EA_max: Optional[float] = Query(
            None, description="Maximum value for the electron affinity in eV.",
        ),
        EA_min: Optional[float] = Query(
            None, description="Minimum value for the electron affinity in eV.",
        ),
        IE_max: Optional[float] = Query(
            None, description="Maximum value for the ionization energy in eV.",
        ),
        IE_min: Optional[float] = Query(
            None, description="Minimum value for the ionization energy in eV.",
        ),
        charge_max: Optional[int] = Query(
            None, description="Maximum value for the charge in +e.",
        ),
        charge_min: Optional[int] = Query(
            None, description="Minimum value for the charge in +e.",
        ),
        pointgroup: Optional[str] = Query(
            None, description="Point of the molecule in Schoenflies notation.",
        ),
        smiles: Optional[str] = Query(
            None,
            description="The simplified molecular input line-entry system (SMILES) \
            representation of the molecule.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "nelements": [nelements_min, nelements_max],
            "EA": [EA_min, EA_max],
            "IE": [IE_min, IE_max],
            "charge": [charge_min, charge_max],
        }  # type: dict

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        if pointgroup:
            crit["pointgroup"] = pointgroup

        if smiles:
            crit["smiles"] = smiles

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = self._keys_from_query()
        indexes = []
        for key in keys:
            if "_min" in key:
                key = key.replace("_min", "")
                indexes.append((key, False))
        return indexes


class MoleculeFormulaQuery(QueryOperator):
    """
    Method to generate a query for molecule data using a chemical formula
    """

    def query(
        self,
        formula: Optional[str] = Query(
            None,
            description="Chemical formula of the molecule. A comma-separated list of formulas is also accepted.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if formula:
            formula_list = [f.strip() for f in formula.split(",")]

            try:

                if len(formula_list) == 1:
                    reduced_formula = Composition(
                        formula
                    ).get_reduced_formula_and_factor()[0]
                    crit["formula_pretty"] = reduced_formula
                else:
                    crit["formula_pretty"] = {
                        "$in": [
                            Composition(f).get_reduced_formula_and_factor()[0]
                            for f in formula_list
                        ]
                    }
            except ValueError:
                raise HTTPException(
                    status_code=400,
                    detail="Problem processing one or more provided formulas.",
                )

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("formula_pretty", False)]



from maggma.api.resource import ReadOnlyResource
from emmet.core.polar import PiezoelectricDoc

from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from emmet.api.routes.piezo.query_operators import PiezoelectricQuery
from emmet.api.routes.materials.query_operators import MultiMaterialIDQuery
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings


def piezo_resource(piezo_store):
    resource = ReadOnlyResource(
        piezo_store,
        PiezoelectricDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            PiezoelectricQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(PiezoelectricDoc, default_fields=["material_id", "last_updated"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Piezoelectric"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from typing import Optional, List, Dict, Union, Any
from fastapi import Query
from pymatgen.core import Composition
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS
from emmet.core.synthesis import SynthesisTypeEnum, OperationTypeEnum
from emmet.api.routes.synthesis.utils import mask_highlights, mask_paragraphs


class SynthesisSearchQuery(QueryOperator):
    """
    Method to generate a synthesis text search query
    """

    def query(  # noqa: C901
        self,
        keywords: Optional[str] = Query(
            None,
            description="Comma delimited string keywords to search synthesis paragraph text with.",
        ),
        synthesis_type: Optional[List[SynthesisTypeEnum]] = Query(
            None, description="Type of synthesis to include."
        ),
        target_formula: Optional[str] = Query(
            None, description="Chemical formula of the target material."
        ),
        precursor_formula: Optional[str] = Query(
            None, description="Chemical formula of the precursor material."
        ),
        operations: Optional[List[OperationTypeEnum]] = Query(
            None, description="List of operations that syntheses must have."
        ),
        condition_heating_temperature_min: Optional[float] = Query(
            None, description="Minimal heating temperature."
        ),
        condition_heating_temperature_max: Optional[float] = Query(
            None, description="Maximal heating temperature."
        ),
        condition_heating_time_min: Optional[float] = Query(
            None, description="Minimal heating time."
        ),
        condition_heating_time_max: Optional[float] = Query(
            None, description="Maximal heating time."
        ),
        condition_heating_atmosphere: Optional[List[str]] = Query(
            None, description='Required heating atmosphere, such as "air", "argon".'
        ),
        condition_mixing_device: Optional[List[str]] = Query(
            None, description='Required mixing device, such as "zirconia", "Al2O3".'
        ),
        condition_mixing_media: Optional[List[str]] = Query(
            None, description='Required mixing media, such as "alcohol", "water".'
        ),
        _skip: int = Query(0, description="Number of entries to skip in the search"),
        _limit: int = Query(
            10,
            description="Max number of entries to return in a single query. Limited to 10.",
        ),
    ):
        project_dict: Dict[str, Union[Dict, int]] = {
            "_id": 0,
            "doi": 1,
            "synthesis_type": 1,
            "reaction": 1,
            "reaction_string": 1,
            "operations": 1,
            "target": 1,
            "targets_formula": 1,
            "targets_formula_s": 1,
            "precursors": 1,
            "precursors_formula_s": 1,
            "paragraph_string": 1,
        }

        pipeline: List[Any] = []

        if keywords:
            pipeline.insert(
                0,
                {
                    "$search": {
                        # NOTICE to MongoDB admin:
                        # This MongoDB Atlas index should be created with:
                        # {"mappings": { "dynamic": false,
                        #   "fields": {
                        #     "paragraph_string": {"type": "string"}
                        #   }}}
                        "index": "synth_descriptions",
                        "search": {
                            "query": [
                                keyword.strip() for keyword in keywords.split(",")
                            ],
                            "path": "paragraph_string",
                        },
                        "highlight": {
                            "path": "paragraph_string",
                            # "maxNumPassages": 1
                        },
                    }
                },
            )
            project_dict.update(
                {
                    "search_score": {"$meta": "searchScore"},
                    "highlights": {"$meta": "searchHighlights"},
                }
            )

        pipeline.append({"$project": project_dict})

        crit: Dict[str, Any] = {}

        if synthesis_type:
            crit["synthesis_type"] = {"$in": synthesis_type}
        if target_formula:
            reduced_formula = Composition(
                target_formula
            ).get_reduced_formula_and_factor()[0]
            crit["targets_formula_s"] = reduced_formula
        if precursor_formula:
            reduced_formula = Composition(
                precursor_formula
            ).get_reduced_formula_and_factor()[0]
            crit["precursors_formula_s"] = reduced_formula
        if operations:
            crit["operations.type"] = {"$all": operations}
        if condition_heating_temperature_min is not None:
            field = "operations.conditions.heating_temperature.values"
            if field not in crit:
                crit[field] = {"$elemMatch": {}}
            crit[field]["$elemMatch"]["$gte"] = condition_heating_temperature_min
        if condition_heating_temperature_max is not None:
            field = "operations.conditions.heating_temperature.values"
            if field not in crit:
                crit[field] = {"$elemMatch": {}}
            crit[field]["$elemMatch"]["$lte"] = condition_heating_temperature_max
        if condition_heating_time_min is not None:
            field = "operations.conditions.heating_time.values"
            if field not in crit:
                crit[field] = {"$elemMatch": {}}
            crit[field]["$elemMatch"]["$gte"] = condition_heating_time_min
        if condition_heating_time_max is not None:
            field = "operations.conditions.heating_time.values"
            if field not in crit:
                crit[field] = {"$elemMatch": {}}
            crit[field]["$elemMatch"]["$lte"] = condition_heating_time_max
        if condition_heating_atmosphere:
            crit["operations.conditions.heating_atmosphere"] = {
                "$all": condition_heating_atmosphere
            }
        if condition_mixing_device:
            crit["operations.conditions.mixing_device"] = {
                "$all": condition_mixing_device
            }
        if condition_mixing_media:
            crit["operations.conditions.mixing_media"] = {
                "$all": condition_mixing_media
            }

        if crit:
            if keywords:
                pipeline.insert(2, {"$match": crit})
            else:
                pipeline.insert(0, {"$match": crit})

        pipeline.append(
            {
                "$facet": {
                    "results": [{"$skip": _skip}, {"$limit": _limit}]
                    if not keywords
                    else [
                        {"$sort": {"search_score": -1}},
                        {"$skip": _skip},
                        {"$limit": _limit},
                    ],
                    "total_doc": [{"$count": "count"}],
                }
            }
        )

        pipeline.extend(
            [
                {"$unwind": "$results"},
                {"$unwind": "$total_doc"},
                {
                    "$replaceRoot": {
                        "newRoot": {
                            "$mergeObjects": [
                                "$results",
                                {"total_doc": "$total_doc.count"},
                            ]
                        }
                    }
                },
            ]
        )

        return {"pipeline": pipeline}

    def post_process(self, docs, query):

        self.total_doc = 0

        if len(docs) > 0:
            self.total_doc = docs[0]["total_doc"]

        for doc in docs:
            mask_highlights(doc)
            mask_paragraphs(doc)

        return docs

    def meta(self):
        return {"total_doc": self.total_doc}



from fastapi import Query
from maggma.api.query_operator.core import QueryOperator
from maggma.api.utils import STORE_PARAMS
from emmet.core.mpid import MPID


class ChgcarTaskIDQuery(QueryOperator):
    """
    Method to generate a query on CHGCAR metadata with calculation (task) ID
    """

    def query(
        self,
        task_ids: str = Query(
            None,
            description="Comma-separated list of calculation (task) IDs to query on",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if task_ids:
            crit.update({"task_id": {"$in": [task_id.strip() for task_id in task_ids.split(",")]}})

        return {"criteria": crit}



import os

from emmet.api.core.api import MAPI
from emmet.api.core.settings import MAPISettings
from maggma.stores import MongoURIStore, S3Store

resources = {}

default_settings = MAPISettings()

db_uri = os.environ.get("MPCONTRIBS_MONGO_HOST", None)
db_version = default_settings.DB_VERSION
db_suffix = os.environ["MAPI_DB_NAME_SUFFIX"]
debug = default_settings.DEBUG

# allow db_uri to be set with a different protocol scheme
# but prepend with mongodb+srv:// if not otherwise specified
if len(db_uri.split("://", 1)) < 2:
    db_uri = "mongodb+srv://" + db_uri

if db_uri:

    materials_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="materials",
    )

    bonds_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="bonds",
    )

    formula_autocomplete_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="_id",
        collection_name="formula_autocomplete",
    )

    task_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="task_id",
        collection_name="tasks",
    )

    thermo_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="thermo",
    )

    phase_diagram_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="chemsys",
        collection_name="phase_diagram",
    )

    dielectric_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="dielectric",
    )

    piezoelectric_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="piezoelectric",
    )

    magnetism_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="magnetism",
    )

    phonon_bs_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="material_id",
        collection_name="pmg_ph_bs",
    )

    eos_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="task_id",
        collection_name="eos",
    )

    similarity_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="material_id",
        collection_name="similarity",
    )

    xas_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="spectrum_id",
        collection_name="xas",
    )

    gb_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="task_id",
        collection_name="grain_boundaries",
    )

    fermi_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="task_id",
        collection_name="fermi_surface",
    )

    elasticity_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="task_id",
        collection_name="elasticity",
    )

    doi_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="task_id",
        collection_name="dois",
    )

    substrates_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="film_id",
        collection_name="substrates",
    )

    surface_props_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="task_id",
        collection_name="surface_properties",
    )

    robo_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="robocrys",
    )

    synth_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="_id",
        collection_name="synth_descriptions",
    )

    insertion_electrodes_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="battery_id",
        collection_name="insertion_electrodes",
    )

    molecules_store = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="task_id",
        collection_name="molecules",
    )

    oxi_states_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="oxi_states",
    )

    provenance_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="provenance",
    )

    alloy_pairs_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="pair_id",
        collection_name="alloy_pairs",
    )

    summary_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="summary",
    )

    es_store = MongoURIStore(
        uri=db_uri,
        database=f"mp_core_{db_suffix}",
        key="material_id",
        collection_name="electronic_structure",
    )

    s3_bs_index = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="fs_id",
        collection_name="s3_bandstructure_index",
    )

    s3_dos_index = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="fs_id",
        collection_name="s3_dos_index",
    )

    s3_bs = S3Store(
        index=s3_bs_index,
        bucket="mp-bandstructures",
        compress=True,
        key="fs_id",
        unpack_data=False,
        searchable_fields=["task_id", "fs_id"],
    )

    s3_dos = S3Store(
        index=s3_dos_index,
        bucket="mp-dos",
        compress=True,
        key="fs_id",
        unpack_data=False,
        searchable_fields=["task_id", "fs_id"],
    )

    s3_chgcar_index = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="fs_id",
        collection_name="atomate_chgcar_fs_index",
    )

    s3_chgcar = S3Store(
        index=s3_chgcar_index,
        bucket="mp-volumetric",
        sub_dir="atomate_chgcar_fs/",
        compress=True,
        key="fs_id",
        unpack_data=False,
        searchable_fields=["task_id", "fs_id"],
    )

    chgcar_url = MongoURIStore(
        uri=db_uri,
        database="mp_core",
        key="fs_id",
        collection_name="chgcar_s3_urls",
    )

    mpcomplete_store = MongoURIStore(
        uri=db_uri,
        database="mp_consumers",
        key="submission_id",
        collection_name="mpcomplete",
    )

    consumer_settings_store = MongoURIStore(
        uri=db_uri,
        database="mp_consumers",
        key="consumer_id",
        collection_name="settings",
    )

    general_store = MongoURIStore(
        uri=db_uri,
        database="mp_consumers",
        key="submission_id",
        collection_name="general_store",
    )
else:
    raise RuntimeError("Must specify MongoDB URI containing inputs.")

# Materials
from emmet.api.routes.materials.resources import (
    materials_resource,
    find_structure_resource,
    formula_autocomplete_resource,
)

resources.update(
    {
        "materials": [
            find_structure_resource(materials_store),
            formula_autocomplete_resource(formula_autocomplete_store),
            materials_resource(materials_store),
        ]
    }
)

# Bonds
from emmet.api.routes.bonds.resources import bonds_resource

resources.update({"bonds": [bonds_resource(bonds_store)]})

# Tasks
from emmet.api.routes.tasks.resources import (
    task_resource,
    trajectory_resource,
    entries_resource,
    task_deprecation_resource,
)

resources.update(
    {
        "tasks": [
            trajectory_resource(task_store),
            entries_resource(task_store),
            task_deprecation_resource(materials_store),
            task_resource(task_store),
        ]
    }
)

# Thermo
from emmet.api.routes.thermo.resources import phase_diagram_resource, thermo_resource

resources.update(
    {
        "thermo": [
            phase_diagram_resource(phase_diagram_store),
            thermo_resource(thermo_store),
        ]
    }
)

# Dielectric
from emmet.api.routes.dielectric.resources import dielectric_resource

resources.update({"dielectric": [dielectric_resource(dielectric_store)]})

# Piezoelectric
from emmet.api.routes.piezo.resources import piezo_resource

resources.update({"piezoelectric": [piezo_resource(piezoelectric_store)]})

# Magnetism
from emmet.api.routes.magnetism.resources import magnetism_resource

resources.update({"magnetism": [magnetism_resource(magnetism_store)]})

# Phonon
from emmet.api.routes.phonon.resources import phonon_bsdos_resource

resources.update({"phonon": [phonon_bsdos_resource(phonon_bs_store)]})

# EOS
from emmet.api.routes.eos.resources import eos_resource

resources.update({"eos": [eos_resource(eos_store)]})

# Similarity
from emmet.api.routes.similarity.resources import similarity_resource

resources.update({"similarity": [similarity_resource(similarity_store)]})

# XAS
from emmet.api.routes.xas.resources import xas_resource

resources.update({"xas": [xas_resource(xas_store)]})

# Grain Boundaries
from emmet.api.routes.grain_boundary.resources import gb_resource

resources.update({"grain_boundary": [gb_resource(gb_store)]})

# Fermi Surface
from emmet.api.routes.fermi.resources import fermi_resource

resources.update({"fermi": [fermi_resource(fermi_store)]})

# Elasticity
from emmet.api.routes.elasticity.resources import elasticity_resource

resources.update({"elasticity": [elasticity_resource(elasticity_store)]})

# DOIs
from emmet.api.routes.dois.resources import dois_resource

resources.update({"doi": [dois_resource(doi_store)]})

# Substrates
from emmet.api.routes.substrates.resources import substrates_resource

resources.update({"substrates": [substrates_resource(substrates_store)]})

# Surface Properties
from emmet.api.routes.surface_properties.resources import surface_props_resource

resources.update({"surface_properties": [surface_props_resource(surface_props_store)]})


# Robocrystallographer
from emmet.api.routes.robocrys.resources import robo_resource, robo_search_resource

resources.update(
    {"robocrys": [robo_search_resource(robo_store), robo_resource(robo_store)]}
)

# Synthesis
from emmet.api.routes.synthesis.resources import synth_resource

resources.update({"synthesis": [synth_resource(synth_store)]})

# Electrodes
from emmet.api.routes.electrodes.resources import insertion_electrodes_resource

resources.update(
    {
        "insertion_electrodes": [
            insertion_electrodes_resource(insertion_electrodes_store)
        ]
    }
)

# Molecules
from emmet.api.routes.molecules.resources import molecules_resource

resources.update({"molecules": [molecules_resource(molecules_store)]})

# Oxidation States
from emmet.api.routes.oxidation_states.resources import oxi_states_resource

resources.update({"oxidation_states": [oxi_states_resource(oxi_states_store)]})

# Alloys
from emmet.api.routes.alloys.resources import alloy_pairs_resource

resources.update({"alloys": [alloy_pairs_resource(alloy_pairs_store)]})

# Provenance
from emmet.api.routes.provenance.resources import provenance_resource

resources.update({"provenance": [provenance_resource(provenance_store)]})

# Charge Density
from emmet.api.routes.charge_density.resources import (
    charge_density_resource,
    charge_density_url_resource,
)

resources.update(
    {
        "charge_density": [
            charge_density_resource(s3_chgcar),
            charge_density_url_resource(chgcar_url),
        ]
    }
)

# Summary
from emmet.api.routes.summary.resources import summary_resource, summary_stats_resource

resources.update(
    {
        "summary": [
            summary_stats_resource(summary_store),
            summary_resource(summary_store),
        ]
    }
)

# Electronic Structure
from emmet.api.routes.electronic_structure.resources import (
    es_resource,
    bs_resource,
    bs_obj_resource,
    dos_resource,
    dos_obj_resource,
)

resources.update(
    {
        "electronic_structure": [
            bs_resource(es_store),
            dos_resource(es_store),
            es_resource(es_store),
            bs_obj_resource(s3_bs),
            dos_obj_resource(s3_dos),
        ]
    }
)
# MPComplete
from emmet.api.routes.mpcomplete.resources import mpcomplete_resource

resources.update({"mpcomplete": [mpcomplete_resource(mpcomplete_store)]})

# Consumers
from emmet.api.routes._consumer.resources import settings_resource

resources.update({"_user_settings": [settings_resource(consumer_settings_store)]})

# General Store
from emmet.api.routes._general_store.resources import general_store_resource

resources.update({"_general_store": [general_store_resource(general_store)]})

# === MAPI setup
from emmet.api.core.documentation import description, tags_meta

api = MAPI(
    resources=resources, debug=debug, description=description, tags_meta=tags_meta
)
app = api.app



from maggma.api.query_operator import (
    NumericQuery,
    PaginationQuery,
    SortQuery,
    SparseFieldsQuery,
    StringQueryOperator,
)
from maggma.api.resource import ReadOnlyResource

from emmet.api.routes.substrates.query_operators import SubstrateStructureQuery
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.core.substrates import SubstratesDoc
from emmet.api.core.settings import MAPISettings


def substrates_resource(substrates_store):
    resource = ReadOnlyResource(
        substrates_store,
        SubstratesDoc,
        query_operators=[
            SubstrateStructureQuery(),
            NumericQuery(model=SubstratesDoc),
            StringQueryOperator(
                model=SubstratesDoc, excluded_fields=["film_orient", "orient"]
            ),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(SubstratesDoc, default_fields=["film_id", "sub_id"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Substrates"],
        enable_get_by_key=False,
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT
    )

    return resource



from maggma.api.resource import ReadOnlyResource
from emmet.core.oxidation_states import OxidationStateDoc

from maggma.api.query_operator import (
    PaginationQuery,
    SortQuery,
    SparseFieldsQuery,
)

from emmet.api.routes.materials.query_operators import FormulaQuery, ChemsysQuery, MultiMaterialIDQuery
from emmet.api.routes.oxidation_states.query_operators import PossibleOxiStateQuery
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings


def oxi_states_resource(oxi_states_store):
    resource = ReadOnlyResource(
        oxi_states_store,
        OxidationStateDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            FormulaQuery(),
            ChemsysQuery(),
            PossibleOxiStateQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(OxidationStateDoc, default_fields=["material_id", "last_updated"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Oxidation States"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS
from emmet.api.routes.electrodes.utils import (
    electrodes_formula_to_criteria,
    electrodes_chemsys_to_criteria,
)
from pymatgen.core.periodic_table import Element
from fastapi import HTTPException

from collections import defaultdict


class ElectrodeFormulaQuery(QueryOperator):
    """
    Method to generate a query for framework formula data
    """

    def query(
        self,
        formula: Optional[str] = Query(
            None,
            description="Query by formula including anonymized formula or by including wild cards. \
A comma delimited string list of anonymous formulas or regular formulas can also be provided.",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if formula:
            crit.update(electrodes_formula_to_criteria(formula))

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [
            ("eentries_composition_summary.all_composition_reduced", False),
            ("entries_composition_summary.all_formulas", False),
        ]


class ElectrodesChemsysQuery(QueryOperator):
    """
    Factory method to generate a dependency for querying by
        chemical system with wild cards.
    """

    def query(
        self,
        chemsys: Optional[str] = Query(
            None,
            description="A comma delimited string list of chemical systems. \
Wildcards for unknown elements only supported for single chemsys queries",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if chemsys:
            crit.update(electrodes_chemsys_to_criteria(chemsys))

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = [
            "entries_composition_summary.all_chemsys",
            "entries_composition_summary.all_elements",
            "nelements",
        ]
        return [(key, False) for key in keys]


class ElectrodeElementsQuery(QueryOperator):
    """
    Factory method to generate a dependency for querying by electrode element data
    """

    def query(
        self,
        elements: Optional[str] = Query(
            None, description="Query by elements in the material composition as a comma-separated list",
        ),
        exclude_elements: Optional[str] = Query(
            None, description="Query by excluded elements in the material composition as a comma-separated list",
        ),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if elements or exclude_elements:
            crit["entries_composition_summary.all_elements"] = {}

        if elements:
            try:
                element_list = [Element(e) for e in elements.strip().split(",")]
            except ValueError:
                raise HTTPException(
                    status_code=400, detail="Problem processing one or more provided elements.",
                )
            crit["entries_composition_summary.all_elements"]["$all"] = [str(el) for el in element_list]

        if exclude_elements:
            try:
                element_list = [Element(e) for e in exclude_elements.strip().split(",")]
            except ValueError:
                raise HTTPException(
                    status_code=400, detail="Problem processing one or more provided elements.",
                )
            crit["entries_composition_summary.all_elements"]["$nin"] = [str(el) for el in element_list]

        return {"criteria": crit}


class VoltageStepQuery(QueryOperator):
    """
    Method to generate a query for ranges of voltage step data values
    """

    def query(
        self,
        delta_volume_max: Optional[float] = Query(
            None, description="Maximum value for the max volume change in percent for a particular voltage step.",
        ),
        delta_volume_min: Optional[float] = Query(
            None, description="Minimum value for the max volume change in percent for a particular voltage step.",
        ),
        average_voltage_max: Optional[float] = Query(
            None, description="Maximum value for the average voltage for a particular voltage step in V.",
        ),
        average_voltage_min: Optional[float] = Query(
            None, description="Minimum value for the average voltage for a particular voltage step in V.",
        ),
        max_voltage_max: Optional[float] = Query(
            None, description="Maximum value for the maximum voltage for a particular voltage step in V.",
        ),
        max_voltage_min: Optional[float] = Query(
            None, description="Minimum value for the maximum voltage for a particular voltage step in V.",
        ),
        min_voltage_max: Optional[float] = Query(
            None, description="Maximum value for the minimum voltage for a particular voltage step in V.",
        ),
        min_voltage_min: Optional[float] = Query(
            None, description="Minimum value for the minimum voltage for a particular voltage step in V.",
        ),
        capacity_grav_max: Optional[float] = Query(
            None, description="Maximum value for the gravimetric capacity in maH/g.",
        ),
        capacity_grav_min: Optional[float] = Query(
            None, description="Minimum value for the gravimetric capacity in maH/g.",
        ),
        capacity_vol_max: Optional[float] = Query(
            None, description="Maximum value for the volumetric capacity in maH/cc.",
        ),
        capacity_vol_min: Optional[float] = Query(
            None, description="Minimum value for the volumetric capacity in maH/cc.",
        ),
        energy_grav_max: Optional[float] = Query(
            None, description="Maximum value for the gravimetric energy (specific energy) in Wh/kg.",
        ),
        energy_grav_min: Optional[float] = Query(
            None, description="Minimum value for the gravimetric energy (specific energy) in Wh/kg.",
        ),
        energy_vol_max: Optional[float] = Query(
            None, description="Maximum value for the volumetric energy (energy_density) in Wh/l.",
        ),
        energy_vol_min: Optional[float] = Query(
            None, description="Minimum value for the volumetric energy (energy_density) in Wh/l.",
        ),
        fracA_charge_max: Optional[float] = Query(
            None, description="Maximum value for the atomic fraction of the working ion in the charged state.",
        ),
        fracA_charge_min: Optional[float] = Query(
            None, description="Minimum value for the atomic fraction of the working ion in the charged state.",
        ),
        fracA_discharge_max: Optional[float] = Query(
            None, description="Maximum value for the atomic fraction of the working ion in the discharged state.",
        ),
        fracA_discharge_min: Optional[float] = Query(
            None, description="Minimum value for the atomic fraction of the working ion in the discharged state.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "max_delta_volume": [delta_volume_min, delta_volume_max],
            "average_voltage": [average_voltage_min, average_voltage_max],
            "max_voltage": [max_voltage_min, max_voltage_max],
            "min_voltage": [min_voltage_min, min_voltage_max],
            "capacity_grav": [capacity_grav_min, capacity_grav_max],
            "capacity_vol": [capacity_vol_min, capacity_vol_max],
            "energy_grav": [energy_grav_min, energy_grav_max],
            "energy_vol": [energy_vol_min, energy_vol_max],
            "fracA_charge": [fracA_charge_min, fracA_charge_max],
            "fracA_discharge": [fracA_discharge_min, fracA_discharge_max],
        }

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = [key for key in self._keys_from_query() if key != "delta_volume_min"]
        indexes = []
        for key in keys:
            if "_min" in key:
                key = key.replace("_min", "")
                indexes.append((key, False))
        indexes.append(("max_delta_volume", False))
        return indexes


class InsertionVoltageStepQuery(QueryOperator):
    """
    Method to generate a query for ranges of insertion voltage step data values
    """

    def query(
        self,
        stability_charge_max: Optional[float] = Query(
            None, description="The maximum value of the energy above hull of the charged material.",
        ),
        stability_charge_min: Optional[float] = Query(
            None, description="The minimum value of the energy above hull of the charged material.",
        ),
        stability_discharge_max: Optional[float] = Query(
            None, description="The maximum value of the energy above hull of the discharged material.",
        ),
        stability_discharge_min: Optional[float] = Query(
            None, description="The minimum value of the energy above hull of the discharged material.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "stability_charge": [stability_charge_min, stability_charge_max],
            "stability_discharge": [stability_discharge_min, stability_discharge_max],
        }

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = self._keys_from_query()

        indexes = []
        for key in keys:
            if "_min" in key:
                key = key.replace("_min", "")
                indexes.append((key, False))
        return indexes


class WorkingIonQuery(QueryOperator):
    """
    Method to generate a query for ranges of insertion electrode data values
    """

    def query(
        self,
        working_ion: Optional[str] = Query(
            None, title="Element of the working ion, or comma-delimited string list of working ion elements.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if working_ion:
            element_list = [element.strip() for element in working_ion.split(",")]
            if len(element_list) == 1:
                crit["working_ion"] = element_list[0]
            else:
                crit["working_ion"] = {"$in": element_list}

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("working_ion", False)]


class ElectrodeMultiMaterialIDQuery(QueryOperator):
    """
    Method to generate a query for different root-level material_id values
    """

    def query(
        self,
        material_ids: Optional[str] = Query(None, description="Comma-separated list of material_id values to query on"),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if material_ids:
            crit.update({"material_ids": {"$in": [material_id.strip() for material_id in material_ids.split(",")]}})

        return {"criteria": crit}


class MultiBatteryIDQuery(QueryOperator):
    """
    Method to generate a query for different root-level battery_id values
    """

    def query(
        self,
        battery_ids: Optional[str] = Query(None, description="Comma-separated list of battery_id values to query on"),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if battery_ids:

            battery_id_list = [material_id.strip() for material_id in battery_ids.split(",")]

            if len(battery_id_list) == 1:
                crit.update({"battery_id": battery_id_list[0]})
            else:
                crit.update({"battery_id": {"$in": battery_id_list}})

        return {"criteria": crit}



from email import header
from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from maggma.api.query_operator.dynamic import NumericQuery
from maggma.api.resource import ReadOnlyResource
from emmet.api.core.settings import MAPISettings

from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.routes.surface_properties.query_operators import ReconstructedQuery
from emmet.core.surface_properties import SurfacePropDoc


def surface_props_resource(surface_prop_store):
    resource = ReadOnlyResource(
        surface_prop_store,
        SurfacePropDoc,
        query_operators=[
            NumericQuery(model=SurfacePropDoc),
            ReconstructedQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(SurfacePropDoc, default_fields=["task_id"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Surface Properties"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT,
    )

    return resource



from fastapi import HTTPException, Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS


class RoboTextSearchQuery(QueryOperator):
    """
    Method to generate a robocrystallographer text search query
    """

    def query(
        self,
        keywords: str = Query(
            ...,
            description="Comma delimited string keywords to search robocrystallographer description text with",
        ),
        _skip: int = Query(0, description="Number of entries to skip in the search"),
        _limit: int = Query(
            100,
            description="Max number of entries to return in a single query. Limited to 100",
        ),
    ) -> STORE_PARAMS:

        if not keywords.strip():
            raise HTTPException(
                status_code=400, detail="Must provide search keywords.",
            )

        pipeline = [
            {
                "$search": {
                    "index": "description",
                    "regex": {
                        "query": [word.strip() for word in keywords.split(",") if word],
                        "path": "description",
                        "allowAnalyzedField": True,
                    },
                }
            },
            {
                "$facet": {
                    "total_doc": [{"$count": "count"}],
                    "results": [
                        {
                            "$project": {
                                "_id": 0,
                                "task_id": 1,
                                "description": 1,
                                "condensed_structure": 1,
                                "last_updates": 1,
                                "search_score": {"$meta": "searchScore"},
                            }
                        }
                    ],
                }
            },
            {"$unwind": "$results"},
            {"$unwind": "$total_doc"},
            {
                "$replaceRoot": {
                    "newRoot": {
                        "$mergeObjects": ["$results", {"total_doc": "$total_doc.count"}]
                    }
                }
            },
            {"$sort": {"search_score": -1}},
            {"$skip": _skip},
            {"$limit": _limit},
        ]
        return {"pipeline": pipeline}

    def post_process(self, docs, query):
        self.total_doc = docs[0]["total_doc"]
        return docs

    def meta(self):
        return {"total_doc": self.total_doc}

    def ensure_indexes(self):  # pragma: no cover
        return [("description", False)]



from emmet.core.summary import SummaryDoc

from maggma.api.query_operator import (
    PaginationQuery,
    SortQuery,
    SparseFieldsQuery,
    NumericQuery,
)
from maggma.api.resource import ReadOnlyResource, AggregationResource
from emmet.api.routes.materials.query_operators import (
    DeprecationQuery,
    ElementsQuery,
    FormulaQuery,
    ChemsysQuery,
    SymmetryQuery,
)
from emmet.api.routes.oxidation_states.query_operators import PossibleOxiStateQuery
from emmet.core.summary import SummaryStats
from emmet.api.routes.summary.hint_scheme import SummaryHintScheme
from emmet.api.routes.summary.query_operators import (
    HasPropsQuery,
    MaterialIDsSearchQuery,
    SearchIsStableQuery,
    SearchIsTheoreticalQuery,
    SearchMagneticQuery,
    SearchHasReconstructedQuery,
    SearchStatsQuery,
    SearchESQuery,
)

from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings

timeout = MAPISettings().TIMEOUT


def summary_resource(summary_store):
    resource = ReadOnlyResource(
        summary_store,
        SummaryDoc,
        query_operators=[
            MaterialIDsSearchQuery(),
            FormulaQuery(),
            ChemsysQuery(),
            ElementsQuery(),
            PossibleOxiStateQuery(),
            SymmetryQuery(),
            SearchIsStableQuery(),
            SearchIsTheoreticalQuery(),
            SearchMagneticQuery(),
            SearchESQuery(),
            NumericQuery(model=SummaryDoc, excluded_fields=["composition"]),
            SearchHasReconstructedQuery(),
            HasPropsQuery(),
            DeprecationQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(SummaryDoc, default_fields=["material_id"]),
        ],
        hint_scheme=SummaryHintScheme(),
        header_processor=GlobalHeaderProcessor(),
        tags=["Summary"],
        disable_validation=True,
        timeout=timeout
    )

    return resource


def summary_stats_resource(summary_store):
    resource = AggregationResource(
        summary_store,
        SummaryStats,
        pipeline_query_operator=SearchStatsQuery(SummaryDoc),
        tags=["Summary"],
        sub_path="/stats/",
        header_processor=GlobalHeaderProcessor(),
        timeout=timeout
    )

    return resource



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS

from collections import defaultdict


class MaterialIDsSearchQuery(QueryOperator):
    """
    Method to generate a query on search docs using multiple material_id values
    """

    def query(
        self,
        material_ids: Optional[str] = Query(
            None, description="Comma-separated list of material_ids to query on"
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if material_ids:

            terminal_search = {
                "_search.id": {
                    "$in": [material_id.strip() for material_id in material_ids.split(",")]
                }
            }

            member_search = {
                "_search.member_ids": {
                    "$in": [material_id.strip() for material_id in material_ids.split(",")]
                }
            }

            crit.update({"$or": [terminal_search, member_search]})

        return {"criteria": crit}


class FormulaSearchQuery(QueryOperator):

    def query(
        self,
        formulae: Optional[str] = Query(
            None, description="Comma-separated list of end-point formulas to query."
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if formulae:

            formula_search = {
                "_search.formula": {
                    "$in": [formula.strip() for formula in formulae.split(",")]
                }
            }

            crit.update(formula_search)

        return {"criteria": crit}



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS

from collections import defaultdict

from pymatgen.analysis.magnetism import Ordering


class MagneticQuery(QueryOperator):
    """
    Method to generate a query for magnetic data.
    """

    def query(
        self,
        ordering: Optional[Ordering] = Query(
            None, description="Magnetic ordering of the material."
        ),
        total_magnetization_max: Optional[float] = Query(
            None, description="Maximum value for the total magnetization.",
        ),
        total_magnetization_min: Optional[float] = Query(
            None, description="Minimum value for the total magnetization.",
        ),
        total_magnetization_normalized_vol_max: Optional[float] = Query(
            None,
            description="Maximum value for the total magnetization normalized with volume.",
        ),
        total_magnetization_normalized_vol_min: Optional[float] = Query(
            None,
            description="Minimum value for the total magnetization normalized with volume.",
        ),
        total_magnetization_normalized_formula_units_max: Optional[float] = Query(
            None,
            description="Maximum value for the total magnetization normalized with formula units.",
        ),
        total_magnetization_normalized_formula_units_min: Optional[float] = Query(
            None,
            description="Minimum value for the total magnetization normalized with formula units.",
        ),
        num_magnetic_sites_max: Optional[int] = Query(
            None, description="Maximum value for the total number of magnetic sites.",
        ),
        num_magnetic_sites_min: Optional[int] = Query(
            None, description="Minimum value for the total number of magnetic sites.",
        ),
        num_unique_magnetic_sites_max: Optional[int] = Query(
            None,
            description="Maximum value for the total number of unique magnetic sites.",
        ),
        num_unique_magnetic_sites_min: Optional[int] = Query(
            None,
            description="Minimum value for the total number of unique magnetic sites.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "total_magnetization": [total_magnetization_min, total_magnetization_max],
            "total_magnetization_normalized_vol": [
                total_magnetization_normalized_vol_min,
                total_magnetization_normalized_vol_max,
            ],
            "total_magnetization_normalized_formula_units": [
                total_magnetization_normalized_formula_units_min,
                total_magnetization_normalized_formula_units_max,
            ],
            "num_magnetic_sites": [num_magnetic_sites_min, num_magnetic_sites_max],
            "num_unique_magnetic_sites": [
                num_unique_magnetic_sites_min,
                num_unique_magnetic_sites_max,
            ],
        }  # type: dict

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        if ordering:
            crit["ordering"] = ordering.value

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = self._keys_from_query()
        indexes = []
        for key in keys:
            if "_min" in key:
                key = key.replace("_min", "")
                indexes.append((key, False))
        return indexes



from maggma.api.resource import HintScheme


class TasksHintScheme(HintScheme):
    """
    Hint scheme for the tasks endpoint.
    """

    def generate_hints(self, query):

        hints = {"hint": {}}

        if query["criteria"] == {}:
            hints["hint"]["_id"] = 1

        return hints



from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from maggma.api.resource import ReadOnlyResource

from emmet.api.routes.materials.query_operators import (
    ChemsysQuery,
    ElementsQuery,
    FormulaQuery,
)
from emmet.api.routes.tasks.hint_scheme import TasksHintScheme
from emmet.api.routes.tasks.query_operators import (
    DeprecationQuery,
    MultipleTaskIDsQuery,
    TrajectoryQuery,
    EntryQuery,
)
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.core.settings import MAPISettings
from emmet.core.tasks import DeprecationDoc, TaskDoc, TrajectoryDoc, EntryDoc

timeout = MAPISettings().TIMEOUT


def task_resource(task_store):
    resource = ReadOnlyResource(
        task_store,
        TaskDoc,
        query_operators=[
            FormulaQuery(),
            ChemsysQuery(),
            ElementsQuery(),
            MultipleTaskIDsQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                TaskDoc, default_fields=["task_id", "formula_pretty", "last_updated"],
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        hint_scheme=TasksHintScheme(),
        tags=["Tasks"],
        timeout=timeout
    )

    return resource


def task_deprecation_resource(materials_store):
    resource = ReadOnlyResource(
        materials_store,
        DeprecationDoc,
        query_operators=[DeprecationQuery(), PaginationQuery()],
        tags=["Tasks"],
        enable_get_by_key=False,
        enable_default_search=True,
        sub_path="/deprecation/",
        header_processor=GlobalHeaderProcessor(),
        timeout=timeout
    )

    return resource


def trajectory_resource(task_store):
    resource = ReadOnlyResource(
        task_store,
        TrajectoryDoc,
        query_operators=[TrajectoryQuery(), PaginationQuery()],
        key_fields=["task_id", "calcs_reversed"],
        tags=["Tasks"],
        sub_path="/trajectory/",
        header_processor=GlobalHeaderProcessor(),
        timeout=timeout
    )

    return resource


def entries_resource(task_store):
    resource = ReadOnlyResource(
        task_store,
        EntryDoc,
        query_operators=[EntryQuery(), PaginationQuery()],
        key_fields=[
            "task_id",
            "input",
            "output",
            "run_type",
            "task_type",
            "completed_at",
            "last_updated",
        ],
        tags=["Tasks"],
        sub_path="/entries/",
        header_processor=GlobalHeaderProcessor(),
        timeout=timeout
    )

    return resource



from maggma.api.resource import ReadOnlyResource
from emmet.core.fermi import FermiDoc
from emmet.api.core.global_header import GlobalHeaderProcessor
from maggma.api.query_operator import PaginationQuery, SparseFieldsQuery


def fermi_resource(fermi_store):
    resource = ReadOnlyResource(
        fermi_store,
        FermiDoc,
        query_operators=[
            PaginationQuery(),
            SparseFieldsQuery(FermiDoc, default_fields=["task_id", "last_updated"]),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Electronic Structure"],
        disable_validation=True,
    )

    return resource



from maggma.api.query_operator.dynamic import NumericQuery
from maggma.api.resource import ReadOnlyResource
from emmet.core.thermo import ThermoDoc
from emmet.core.thermo import PhaseDiagramDoc

from maggma.api.query_operator import (
    PaginationQuery,
    SortQuery,
    SparseFieldsQuery,
)
from emmet.api.routes.thermo.query_operators import IsStableQuery
from emmet.api.core.global_header import GlobalHeaderProcessor
from emmet.api.routes.materials.query_operators import (
    MultiMaterialIDQuery,
    FormulaQuery,
    ChemsysQuery,
)
from emmet.api.core.settings import MAPISettings


def phase_diagram_resource(phase_diagram_store):
    resource = ReadOnlyResource(
        phase_diagram_store,
        PhaseDiagramDoc,
        tags=["Thermo"],
        sub_path="/phase_diagram/",
        disable_validation=True,
        enable_default_search=False,
        header_processor=GlobalHeaderProcessor(),
    )

    return resource


def thermo_resource(thermo_store):
    resource = ReadOnlyResource(
        thermo_store,
        ThermoDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            FormulaQuery(),
            ChemsysQuery(),
            IsStableQuery(),
            NumericQuery(model=ThermoDoc),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                ThermoDoc, default_fields=["material_id", "last_updated"]
            ),
        ],
        header_processor=GlobalHeaderProcessor(),
        tags=["Thermo"],
        disable_validation=True,
        timeout=MAPISettings().TIMEOUT
    )

    return resource



from sys import version_info
from typing import Optional, Union
from fastapi import Query, HTTPException
from pymatgen.analysis.magnetism.analyzer import Ordering
from pymatgen.electronic_structure.core import Spin, OrbitalType
from pymatgen.core.periodic_table import Element
from emmet.core.mpid import MPID

from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS
from emmet.core.electronic_structure import (
    BSPathType,
    DOSProjectionType,
)

from collections import defaultdict

if version_info >= (3, 8):
    from typing import Literal  # type: ignore
else:
    from typing_extensions import Literal  # type: ignore


class ESSummaryDataQuery(QueryOperator):
    """
    Method to generate a query on electronic structure summary data.
    """

    def query(
        self,
        magnetic_ordering: Optional[Ordering] = Query(None, description="Magnetic ordering associated with the data."),
        is_gap_direct: Optional[bool] = Query(None, description="Whether a band gap is direct or not."),
        is_metal: Optional[bool] = Query(None, description="Whether the material is considered a metal."),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if magnetic_ordering:
            crit["magnetic_ordering"] = magnetic_ordering.value

        if is_gap_direct is not None:
            crit["is_gap_direct"] = is_gap_direct

        if is_metal is not None:
            crit["is_metal"] = is_metal

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover

        keys = ["band_gap", "efermi", "magnetic_ordering", "is_gap_direct", "is_metal"]

        return [(key, False) for key in keys]


class BSDataQuery(QueryOperator):
    """
    Method to generate a query on band structure data.
    """

    def query(
        self,
        path_type: Optional[BSPathType] = Query(
            None,
            description="k-path selection convention for the band structure.",
        ),
        band_gap_max: Optional[float] = Query(None, description="Maximum value for the band gap energy in eV."),
        band_gap_min: Optional[float] = Query(None, description="Minimum value for the band gap energy in eV."),
        efermi_max: Optional[float] = Query(None, description="Maximum value for the fermi energy in eV."),
        efermi_min: Optional[float] = Query(None, description="Minimum value for the fermi energy in eV."),
        magnetic_ordering: Optional[Ordering] = Query(None, description="Magnetic ordering associated with the data."),
        is_gap_direct: Optional[bool] = Query(None, description="Whether a band gap is direct or not."),
        is_metal: Optional[bool] = Query(None, description="Whether the material is considered a metal."),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if path_type is not None:

            d = {
                f"bandstructure.{path_type.value}.band_gap": [
                    band_gap_min,
                    band_gap_max,
                ],
                f"bandstructure.{path_type.value}.efermi": [efermi_min, efermi_max],
            }

            for entry in d:
                if d[entry][0] is not None:
                    crit[entry]["$gte"] = d[entry][0]

                if d[entry][1] is not None:
                    crit[entry]["$lte"] = d[entry][1]

            if magnetic_ordering:
                crit[f"bandstructure.{path_type.value}.magnetic_ordering"] = magnetic_ordering.value

            if is_gap_direct is not None:
                crit[f"bandstructure.{path_type.value}.is_gap_direct"] = is_gap_direct

            if is_metal is not None:
                crit[f"bandstructure.{path_type.value}.is_metal"] = is_metal

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover

        keys = ["bandstructure"]

        for bs_type in BSPathType:
            for field in ["band_gap", "efermi"]:
                keys.append(f"bandstructure.{bs_type.value}.{field}")

        return [(key, False) for key in keys]


class DOSDataQuery(QueryOperator):
    """
    Method to generate a query on density of states summary data.
    """

    def query(
        self,
        projection_type: Optional[DOSProjectionType] = Query(
            None,
            description="Projection type for the density of states data.",
        ),
        spin: Optional[Union[Literal["1", "-1"], Spin]] = Query(
            None,
            description="Spin channel for density of states data. '1' corresponds to spin up.",
        ),
        element: Optional[Element] = Query(
            None,
            description="Element type for projected density of states data.",
        ),
        orbital: Optional[OrbitalType] = Query(
            None,
            description="Orbital type for projected density of states data.",
        ),
        band_gap_max: Optional[float] = Query(None, description="Maximum value for the band gap energy in eV."),
        band_gap_min: Optional[float] = Query(None, description="Minimum value for the band gap energy in eV."),
        efermi_max: Optional[float] = Query(None, description="Maximum value for the fermi energy in eV."),
        efermi_min: Optional[float] = Query(None, description="Minimum value for the fermi energy in eV."),
        magnetic_ordering: Optional[Ordering] = Query(None, description="Magnetic ordering associated with the data."),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if isinstance(spin, str):
            spin = Spin(int(spin))

        if projection_type is not None:
            if spin is None:
                raise HTTPException(
                    status_code=400,
                    detail="Must specify a spin channel for querying dos summary data.",
                )
            else:

                d = {
                    "band_gap": [band_gap_min, band_gap_max],
                    "efermi": [efermi_min, efermi_max],
                }

                for entry in d:
                    if projection_type.value == "total":

                        key_prefix = f"total.{str(spin.value)}"

                    elif projection_type.value == "orbital":

                        if orbital is None:
                            raise HTTPException(
                                status_code=400,
                                detail="Must specify an orbital type for querying orbital projection data.",
                            )

                        key_prefix = f"orbital.{str(orbital.name)}.{str(spin.value)}"

                    elif projection_type.value == "elemental":

                        if element is None:
                            raise HTTPException(
                                status_code=400,
                                detail="Must specify an element type for querying element projection data.",
                            )

                        if orbital is not None:
                            key_prefix = f"elemental.{str(element.value)}.{str(orbital.name)}.{str(spin.value)}"

                        else:
                            key_prefix = f"elemental.{str(element.value)}.total.{str(spin.value)}"

                    key = f"dos.{key_prefix}.{entry}"

                    if d[entry][0] is not None:
                        crit[key]["$gte"] = d[entry][0]
                    if d[entry][1] is not None:
                        crit[key]["$lte"] = d[entry][1]

        if magnetic_ordering:
            crit.update({"dos.magnetic_ordering": magnetic_ordering.value})

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover

        keys = ["dos", "dos.magnetic_ordering"]

        for proj_type in DOSProjectionType:
            keys.append(f"dos.{proj_type.value}.$**")

        return [(key, False) for key in keys]


class ObjectQuery(QueryOperator):
    """
    Method to generate a query on object data by task_id.
    """

    def query(
        self,
        task_id: MPID = Query(
            ...,
            description="The calculation (task) ID associated with the data object",
        ),
    ) -> STORE_PARAMS:

        return {"criteria": {"task_id": str(task_id)}}



from maggma.api.resource.core import HeaderProcessor
from fastapi import Response, Request


class GlobalHeaderProcessor(HeaderProcessor):
    def process_header(self, response: Response, request: Request):
        groups = request.headers.get("X-Authenticated-Groups", None)
        if groups is not None and "api_all_nolimit" in [
            group.strip() for group in groups.split(",")
        ]:
            response.headers["X-Bypass-Rate-Limit"] = "ALL"

        # forward Consumer Id header in response
        consumer_id = request.headers.get("X-Consumer-Id", "-")
        response.headers["X-Consumer-Id"] = consumer_id



from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS
from emmet.api.routes.tasks.utils import calcs_reversed_to_trajectory, task_to_entry
from fastapi import Query
from typing import Optional
from monty.json import jsanitize


class MultipleTaskIDsQuery(QueryOperator):
    """
    Method to generate a query on search docs using multiple task_id values
    """

    def query(
        self,
        task_ids: Optional[str] = Query(
            None, description="Comma-separated list of task_ids to query on"
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if task_ids:
            crit.update(
                {
                    "task_id": {
                        "$in": [task_id.strip() for task_id in task_ids.split(",")]
                    }
                }
            )

        return {"criteria": crit}


class TrajectoryQuery(QueryOperator):
    """
    Method to generate a query on calculation trajectory data from task documents
    """

    def query(
        self,
        task_ids: Optional[str] = Query(
            None, description="Comma-separated list of task_ids to query on"
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if task_ids:
            crit.update(
                {
                    "task_id": {
                        "$in": [task_id.strip() for task_id in task_ids.split(",")]
                    }
                }
            )

        return {"criteria": crit}

    def post_process(self, docs, query):
        """
        Post processing to generatore trajectory data
        """

        d = [
            {
                "task_id": doc["task_id"],
                "trajectories": jsanitize(
                    calcs_reversed_to_trajectory(doc["calcs_reversed"])
                ),
            }
            for doc in docs
        ]

        return d


class EntryQuery(QueryOperator):
    """
    Method to generate a query on calculation entry data from task documents
    """

    def query(
        self,
        task_ids: Optional[str] = Query(
            None, description="Comma-separated list of task_ids to query on"
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if task_ids:
            crit.update(
                {
                    "task_id": {
                        "$in": [task_id.strip() for task_id in task_ids.split(",")]
                    }
                }
            )

        return {"criteria": crit}

    def post_process(self, docs, query):
        """
        Post processing to generatore entry data
        """

        d = [
            {"task_id": doc["task_id"], "entry": jsanitize(task_to_entry(doc))}
            for doc in docs
        ]

        return d


class DeprecationQuery(QueryOperator):
    """
    Method to generate a query on calculation trajectory data from task documents
    """

    def query(
        self,
        task_ids: str = Query(
            ..., description="Comma-separated list of task_ids to query on"
        ),
    ) -> STORE_PARAMS:

        self.task_ids = [task_id.strip() for task_id in task_ids.split(",")]

        crit = {}

        if task_ids:
            crit.update({"deprecated_tasks": {"$in": self.task_ids}})

        return {"criteria": crit}

    def post_process(self, docs, query):
        """
        Post processing to generatore deprecation data
        """

        d = []

        for task_id in self.task_ids:
            deprecation = {
                "task_id": task_id,
                "deprecated": False,
                "deprecation_reason": None,
            }
            for doc in docs:
                if task_id in doc["deprecated_tasks"]:
                    deprecation = {
                        "task_id": task_id,
                        "deprecated": True,
                        "deprecation_reason": None,
                    }
                    break

            d.append(deprecation)

        return d



from sys import version_info
from enum import Enum
from typing import Optional
from fastapi import Query

from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS
from emmet.core.summary import SummaryStats

from pymatgen.analysis.magnetism import Ordering

from scipy.stats import gaussian_kde
import numpy as np

from collections import defaultdict

if version_info >= (3, 8):
    from typing import Literal  # type: ignore
else:
    from typing_extensions import Literal  # type: ignore


class HasPropsQuery(QueryOperator):
    """
    Method to generate a query on whether a material has a certain property
    """

    def query(
        self,
        has_props: Optional[str] = Query(
            None,
            description="Comma-delimited list of possible properties given by HasPropsEnum to search for.",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if has_props:
            crit = {
                "has_props": {"$all": [prop.strip() for prop in has_props.split(",")]}
            }

        return {"criteria": crit}


class MaterialIDsSearchQuery(QueryOperator):
    """
    Method to generate a query on search docs using multiple material_id values
    """

    def query(
        self,
        material_ids: Optional[str] = Query(
            None, description="Comma-separated list of material_ids to query on"
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if material_ids:
            crit.update(
                {
                    "material_id": {
                        "$in": [
                            material_id.strip()
                            for material_id in material_ids.split(",")
                        ]
                    }
                }
            )

        return {"criteria": crit}

    def post_process(self, docs, query):

        if not query.get("sort", None):
            mpid_list = (
                query.get("criteria", {}).get("material_id", {}).get("$in", None)
            )

            if mpid_list is not None and "material_id" in query.get("properties", []):
                mpid_mapping = {mpid: ind for ind, mpid in enumerate(mpid_list)}

                docs = sorted(docs, key=lambda d: mpid_mapping[d["material_id"]])

        return docs


class SearchIsStableQuery(QueryOperator):
    """
    Method to generate a query on whether a material is stable
    """

    def query(
        self,
        is_stable: Optional[bool] = Query(
            None, description="Whether the material is stable."
        ),
    ):

        crit = {}

        if is_stable is not None:
            crit["is_stable"] = is_stable

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("is_stable", False)]


class SearchHasReconstructedQuery(QueryOperator):
    """
    Method to generate a query on whether a material has any reconstructed surfaces
    """

    def query(
        self,
        has_reconstructed: Optional[bool] = Query(
            None, description="Whether the material has reconstructed surfaces."
        ),
    ):

        crit = {}

        if has_reconstructed is not None:
            crit["has_reconstructed"] = has_reconstructed

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("is_stable", False)]


class SearchMagneticQuery(QueryOperator):
    """
    Method to generate a query for magnetic data in search docs.
    """

    def query(
        self,
        ordering: Optional[Ordering] = Query(
            None, description="Magnetic ordering of the material."
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if ordering:
            crit["ordering"] = ordering.value

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("ordering", False)]


class SearchIsTheoreticalQuery(QueryOperator):
    """
    Method to generate a query on whether a material is theoretical
    """

    def query(
        self,
        theoretical: Optional[bool] = Query(
            None, description="Whether the material is theoretical."
        ),
    ):

        crit = {}

        if theoretical is not None:
            crit["theoretical"] = theoretical

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        return [("theoretical", False)]


class SearchESQuery(QueryOperator):
    """
    Method to generate a query on search electronic structure data.
    """

    def query(
        self,
        is_gap_direct: Optional[bool] = Query(
            None, description="Whether a band gap is direct or not."
        ),
        is_metal: Optional[bool] = Query(
            None, description="Whether the material is considered a metal."
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if is_gap_direct is not None:
            crit["is_gap_direct"] = is_gap_direct

        if is_metal is not None:
            crit["is_metal"] = is_metal

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover

        keys = ["is_gap_direct", "is_metal"]

        return [(key, False) for key in keys]


class SearchStatsQuery(QueryOperator):
    """
    Method to generate a query on search stats data
    """

    def __init__(self, search_doc):
        valid_numeric_fields = tuple(
            sorted(k for k, v in search_doc.__fields__.items() if v.type_ == float)
        )

        def query(
            field: Literal[valid_numeric_fields] = Query(  # type: ignore
                valid_numeric_fields[0],
                title=f"SearchDoc field to query on, must be a numerical field, "
                f"choose from: {', '.join(valid_numeric_fields)}",
            ),
            num_samples: Optional[int] = Query(
                None, title="If specified, will only sample this number of documents.",
            ),
            min_val: Optional[float] = Query(
                None,
                title="If specified, will only consider documents with field values "
                "greater than or equal to this minimum value.",
            ),
            max_val: Optional[float] = Query(
                None,
                title="If specified, will only consider documents with field values "
                "less than or equal to this minimum value.",
            ),
            num_points: int = Query(
                100, title="The number of values in the returned distribution."
            ),
        ) -> STORE_PARAMS:

            self.num_points = num_points
            self.min_val = min_val
            self.max_val = max_val

            if min_val or max_val:
                pipeline = [{"$match": {field: {}}}]  # type: list
                if min_val is not None:
                    pipeline[0]["$match"][field]["$gte"] = min_val
                if max_val is not None:
                    pipeline[0]["$match"][field]["$lte"] = max_val
            else:
                pipeline = []

            if num_samples:
                pipeline.append({"$sample": {"size": num_samples}})

            pipeline.append({"$project": {field: 1, "_id": 0}})

            return {"pipeline": pipeline}

        self.query = query

    def query(self):
        " Stub query function for abstract class "
        pass

    def post_process(self, docs, query):

        if docs:
            field = list(docs[0].keys())[0]

            num_points = self.num_points
            min_val = self.min_val
            max_val = self.max_val
            num_samples = len(docs)
            warnings = []

            values = [d[field] for d in docs if field in d]
            if min_val is None:
                min_val = min(values)
            if max_val is None:
                max_val = max(values)

            if len(values) != len(docs):
                warnings += [
                    "Some documents have field missing.",
                    f"Only {len(values)} of {len(docs)} ({100*len(values)/len(docs):.2f}%) have {field} field present.",
                ]

            kernel = gaussian_kde(values)

            distribution = list(
                kernel(
                    np.arange(min_val, max_val, step=(max_val - min_val) / num_points,)  # type: ignore
                )
            )

            median = float(np.median(values))
            mean = float(np.mean(values))

            response = SummaryStats(
                field=field,
                num_samples=num_samples,
                min=min_val,
                max=max_val,
                distribution=distribution,
                median=median,
                mean=mean,
                warnings=warnings,
            )

        return [response]


# TODO:
# XAS and GB sub doc query operators
# Add weighted work function to data
# Add dimensionality to search endpoint
# Add "has_reconstructed" data



from typing import Dict
from fastapi import Query, Body
from maggma.api.utils import STORE_PARAMS
from maggma.api.query_operator import QueryOperator


class UserSettingsPostQuery(QueryOperator):
    """Query operators to provide user settings information to post"""

    def query(
        self, consumer_id: str = Query(..., title="Consumer ID",), settings: Dict = Body(..., title="User settings",),
    ) -> STORE_PARAMS:

        crit = {"consumer_id": consumer_id, "settings": settings}

        return {"criteria": crit}

    def post_process(self, docs, query):

        cid = query["criteria"]["consumer_id"]
        settings = query["criteria"]["settings"]

        d = [{"consumer_id": cid, "settings": settings}]

        return d


class UserSettingsGetQuery(QueryOperator):
    """Query operators to provide user settings information"""

    def query(self, consumer_id: str = Query(..., title="Consumer ID",),) -> STORE_PARAMS:

        crit = {"consumer_id": consumer_id}

        return {"criteria": crit}



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS

from collections import defaultdict


class BulkModulusQuery(QueryOperator):
    """
    Method to generate a query for ranges of bulk modulus values
    """

    def query(
        self,
        k_voigt_max: Optional[float] = Query(
            None,
            description="Maximum value for the Voigt average of the bulk modulus in GPa.",
        ),
        k_voigt_min: Optional[float] = Query(
            None,
            description="Minimum value for the Voigt average of the bulk modulus in GPa.",
        ),
        k_reuss_max: Optional[float] = Query(
            None,
            description="Maximum value for the Reuss average of the bulk modulus in GPa.",
        ),
        k_reuss_min: Optional[float] = Query(
            None,
            description="Minimum value for the Reuss average of the bulk modulus in GPa.",
        ),
        k_vrh_max: Optional[float] = Query(
            None,
            description="Maximum value for the Voigt-Reuss-Hill average of the bulk modulus in GPa.",
        ),
        k_vrh_min: Optional[float] = Query(
            None,
            description="Minimum value for the Voigt-Reuss-Hill average of the bulk modulus in GPa.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "elasticity.k_voigt": [k_voigt_min, k_voigt_max],
            "elasticity.k_reuss": [k_reuss_min, k_reuss_max],
            "elasticity.k_vrh": [k_vrh_min, k_vrh_max],
        }

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        return {"criteria": crit}


class ShearModulusQuery(QueryOperator):
    """
    Method to generate a query for ranges of shear modulus values
    """

    def query(
        self,
        g_voigt_max: Optional[float] = Query(
            None,
            description="Maximum value for the Voigt average of the shear modulus in GPa.",
        ),
        g_voigt_min: Optional[float] = Query(
            None,
            description="Minimum value for the Voigt average of the shear modulus in GPa.",
        ),
        g_reuss_max: Optional[float] = Query(
            None,
            description="Maximum value for the Reuss average of the shear modulus in GPa.",
        ),
        g_reuss_min: Optional[float] = Query(
            None,
            description="Minimum value for the Reuss average of the shear modulus in GPa.",
        ),
        g_vrh_max: Optional[float] = Query(
            None,
            description="Maximum value for the Voigt-Reuss-Hill average of the shear modulus in GPa.",
        ),
        g_vrh_min: Optional[float] = Query(
            None,
            description="Minimum value for the Voigt-Reuss-Hill average of the shear modulus in GPa.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "elasticity.g_voigt": [g_voigt_min, g_voigt_max],
            "elasticity.g_reuss": [g_reuss_min, g_reuss_max],
            "elasticity.g_vrh": [g_vrh_min, g_vrh_max],
        }

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        return {"criteria": crit}


class PoissonQuery(QueryOperator):
    """
    Method to generate a query for ranges of
    elastic anisotropy and poisson ratio values
    """

    def query(
        self,
        elastic_anisotropy_max: Optional[float] = Query(
            None, description="Maximum value for the elastic anisotropy.",
        ),
        elastic_anisotropy_min: Optional[float] = Query(
            None, description="Maximum value for the elastic anisotropy.",
        ),
        poisson_max: Optional[float] = Query(
            None, description="Maximum value for Poisson's ratio.",
        ),
        poisson_min: Optional[float] = Query(
            None, description="Minimum value for Poisson's ratio.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        d = {
            "elasticity.universal_anisotropy": [
                elastic_anisotropy_min,
                elastic_anisotropy_max,
            ],
            "elasticity.homogeneous_poisson": [poisson_min, poisson_max],
        }

        for entry in d:
            if d[entry][0] is not None:
                crit[entry]["$gte"] = d[entry][0]

            if d[entry][1] is not None:
                crit[entry]["$lte"] = d[entry][1]

        return {"criteria": crit}


class ElasticityChemsysQuery(QueryOperator):
    """
    Method to generate a query on chemsys data
    """

    def query(
        self,
        chemsys: Optional[str] = Query(
            None, description="A comma delimited string list of chemical systems.",
        ),
    ):

        crit = {}  # type: dict

        if chemsys:
            chemsys_list = [chemsys_val.strip() for chemsys_val in chemsys.split(",")]

            query_vals = []
            for chemsys_val in chemsys_list:
                eles = chemsys_val.split("-")
                sorted_chemsys = "-".join(sorted(eles))
                query_vals.append(sorted_chemsys)

            if len(query_vals) == 1:
                crit["chemsys"] = query_vals[0]
            else:
                crit["chemsys"] = {"$in": query_vals}

        return {"criteria": crit}



from typing import Dict
from fastapi import Query, Body
from maggma.api.utils import STORE_PARAMS
from maggma.api.query_operator import QueryOperator


class GeneralStorePostQuery(QueryOperator):
    """Query operators to provide general store information to post"""

    def query(
        self,
        kind: str = Query(..., title="Data type"),
        markdown: str = Query(None, title="Markdown data"),
        meta: Dict = Body(None, title="Metadata"),
    ) -> STORE_PARAMS:

        crit = {"kind": kind, "markdown": markdown, "meta": meta}

        return {"criteria": crit}

    def post_process(self, docs, query):

        d = [
            {
                "kind": query["criteria"]["kind"],
                "markdown": query["criteria"].get("markdown", None),
                "meta": query["criteria"].get("meta", None),
            }
        ]

        return d


class GeneralStoreGetQuery(QueryOperator):
    """Query operators to obtain general store information"""

    def query(self, kind: str = Query(..., title="Data type")) -> STORE_PARAMS:

        crit = {"kind": kind}

        return {"criteria": crit}



# noqa: E501
description = """
The Materials Project API allows anyone to have direct access to current, up-to-date information from the Materials Project database in a structured way.

This allows for analysis, development of automated tools, machine learning, downloading personal copies of the Materials Project database and more on a large scale.

The API is offered with the hopes of making Materials Project data more useful to you. We want you to use our data! As such, the API is offered free-of-charge and we support several tools to help you get started.

## API Key
To make any request to the Materials Project API, you must use an API key. Your API key is generated for you automatically upon registering with the Materials Project website and is synced with the email you used to register.

Remember to keep your API key safe and to not share it with anyone you do not trust.

If you are logged in, you can always access your API key from this page or from your [dashboard](https://next-gen.materialsproject.org/dashboard).

If you intend heavy API usage, you can give us a heads up by sending a message to <heavy.api.use@materialsproject.org>. With the exception of retrieving charge densities, this is not required, but may help us if we see unusual load on our servers.

## Accessing Data
To use the API, you have three options:

1. You can use our first-party supported Python client. This is the recommend route. The `mp-api` package containing the client is pip installable.

    ```
    pip install mp-api
    ```

    The `MPRester` client can be accessed by importing from it. This will ultimately replace the legacy `MPRester` available in pymatgen.

    For more details on how to use this, including code examples, please see <https://next-gen.materialsproject.org/api>.

2. You can demo the API interactively on this documentation page. Click the "Authorize" button, paste in your API key, and then click the appropriate section to try out a query.

3. Since this is a REST API, and offers a fully-compliant OpenAPI specification, it's possible to use the API with many libraries in many languages and environments, including JavaScript, MATLAB, Mathematica, etc. However, we do not offer first-party support for explaining how to do this, and you will have to follow the specification yourself.

"""

tags_meta = [
    {
        "name": "Summary",
        "description": "Route providing a large amount of amalgamated data for a material. This is constructed by \
            combining subsets of data from many of the other API endpoints. The summary endpoint is very useful for \
            performing queries for materials over a large property space. Note that every unique material within \
            the Materials Project should have a set of summary data. See the `SummaryDoc` schema for a full list of \
            fields returned by this route.",
    },
    {
        "name": "Materials",
        "description": 'Route for "core" information associated with a given material in the Materials Project \
            database. The unique identifier for a material is its `material_id` (e.g. `mp-149`). Core data in \
            this context refers to the crystal structure, information associated with it such as the density \
            and chemical formula, and the associated calculations which are identified with unique `task_id` \
            values. It does not contain any materials properties such as the formation energy or band gap, please \
            consult other property-specific endpoints for this information. See the `MaterialsDoc` schema for \
            a full list of fields returned by this route.',
    },
    {
        "name": "Tasks",
        "description": 'Route for "core" information associated with a given calculation in the Materials Project \
            database. Multiple calculations can ultimately be associated with a unique material, and are the source \
            of its reported properties. The unique identifier for a calculation is its `task_id`. Note \
            that the `material_id` chosen for a given material is sourced from one of the `task_id` values \
            associated with it. Core data in this context refers to calculation quantities such as parsed input \
            and output data (e.g. VASP input flags, atomic forces, structures) and runtime statistics. See the \
            `TaskDoc` schema for a full list of fields returned by this route.',
    },
    {
        "name": "Thermo",
        "description": "Route providing computed thermodynamic data for a material such as \
            formation energy and energy above hull. Corrected energy values are also available that employ \
            the schemes discussed by \
            [Jain *et al.*](https://journals.aps.org/prb/abstract/10.1103/PhysRevB.84.045115) \
            and [Wang *et al.*](https://chemrxiv.org/engage/chemrxiv/article-details/60c758d9469df42a4ef45757)\
            See the `ThermoDoc` schema for a full list of fields returned by this route.",
        "externalDocs": {
            "description": "For a more detailed description",
            "url": "https://docs.materialsproject.org/methodology/total-energies",
        },
    },
    {
        "name": "Dielectric",
        "description": "Route providing computed dielectric data for a material following the \
            methodology discussed by [Petousis *et al.*](https://doi.org/10.1038/sdata.2016.134) \
            Note that dielectric data has not been calculated for all materials in the Materials \
            Project database. See the `DielectricDoc` schema for a full list of fields returned by this route.",
        "externalDocs": {
            "description": "For a more detailed description",
            "url": "https://docs.materialsproject.org/methodology/dielectricity",
        },
    },
    {
        "name": "Magnetism",
        "description": "Route providing computed magnetic ordering related data for a material following the \
            methodology discussed by [Horton *et al.*](https://doi.org/10.1038/s41524-019-0199-7) \
            Note that magnetic data has not been calculated for all materials in the Materials \
            Project database. See the `MagnetismDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Piezoelectric",
        "description": "Route providing computed piezoelectric data for a material following the \
            methodology discussed by [de Jong *et al.*](https://doi.org/10.1038/sdata.2015.53) \
            Note that piezoelectric data has not been calculated for all materials in the Materials \
            Project database. See the `PiezoDoc` schema for a full list of fields returned by this route.",
        "externalDocs": {
            "description": "For a more detailed description",
            "url": "https://docs.materialsproject.org/methodology/piezoelectricity",
        },
    },
    {
        "name": "Phonon",
        "description": "Route providing computed phonon data for a material following the \
            methodology discussed by [Petretto *et al.*](https://doi.org/10.1038/sdata.2018.65) \
            Note that phonon data has not been calculated for all materials in the Materials \
            Project database. See the `PhononBSDOSDoc` schema for a full list of fields returned by this route.",
        "externalDocs": {
            "description": "For a more detailed description",
            "url": "https://docs.materialsproject.org/methodology/phonons",
        },
    },
    {
        "name": "EOS",
        "description": "Route providing computed equations of state data for a material following the \
            methodology discussed by [Latimer *et al.*](https://doi.org/10.1038/s41524-018-0091-x) \
            Note that equations of state data has not been calculated for all materials in the Materials \
            Project database. See the `EOSDoc` schema for a full list of fields returned by this route.",
        "externalDocs": {
            "description": "For a more detailed description",
            "url": "https://docs.materialsproject.org/methodology/equations-of-state",
        },
    },
    {
        "name": "Similarity",
        "description": "Route providing a computed similarity metric between materials following the \
            methodology discussed by Zimmerman *et al.* in [10.3389/fmats.2017.00034](https://doi.org/10.3389/fmats.2017.00034) \
            and [10.1039/C9RA07755C](https://doi.org/10.1039/C9RA07755C). \
            Note that similarity data has not been calculated for all materials in the Materials \
            Project database. See the `imilarityDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "XAS",
        "description": "Route providing computed x-ray absorption spectroscopy data for a material following the \
            methodology discussed by [Mathew *et al.*](https://doi.org/10.1038/sdata.2018.151) \
            and [Chen *et al.*](https://doi.org/10.1038/s41597-021-00936-5) \
            Note that x-ray absorption spectroscopy data has not been calculated for all materials in the Materials \
            Project database. See the `XASDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Grain Boundaries",
        "description": "Route providing computed grain boundary data for a material following the \
            methodology discussed by [Hui *et al.*](https://doi.org/10.1016/j.actamat.2019.12.030) \
            Note that grain boundary data has not been calculated for all materials in the Materials \
            Project database. See the `GrainBoundaryDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Electronic Structure",
        "description": "Routes providing computed electronic structure related data for a material such as \
            band gap and fermi level. Python objects for line-mode band structures, density of states, and \
            fermi surfaces are also available. This data was obtained following the methodology discussed by \
            [Munro *et al.*](https://doi.org/10.1038/s41524-020-00383-7) and [Ganose *et al.*](https://doi.org/10.21105/joss.03089) \
            Note that full band structure, density of states, and fermi surface data has not been calculated for \
            all materials in the Materials Project database. See the `ElectronicStructureDoc` and `FermiDoc` schema \
            for a full list of fields returned by the associated routes.",
        "externalDocs": {
            "description": "For a more detailed description",
            "url": "https://docs.materialsproject.org/methodology/electronic-structure",
        },
    },
    {
        "name": "Elasticity",
        "description": "Route providing computed elasticity data for a material following the \
            methodology discussed by [de Jong *et al.*](https://doi.org/10.1038/sdata.2015.9) \
            Note that elasticity data has not been calculated for all materials in the Materials \
            Project database. See the `ElasticityDoc` schema for a full list of fields returned by this route.",
        "externalDocs": {
            "description": "For a more detailed description",
            "url": "https://docs.materialsproject.org/methodology/elasticity",
        },
    },
    {
        "name": "DOIs",
        "description": "Route providing DOI and bibtex reference information for a material. \
            Note that this data may not be available for all materials in the Materials \
            Project database. See the `DOIDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Substrates",
        "description": "Route providing computed suggested substrate data for a material following the \
            methodology discussed by [Ding *et al.*](https://doi.org/10.1021/acsami.6b01630) \
            Note that substrate data has not been calculated for all materials in the Materials \
            Project database. See the `SubstratesDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Surface Properties",
        "description": "Route providing computed surface property data for a material following the \
            methodology discussed by [Tran *et al.*](https://doi.org/10.1038/sdata.2016.80) \
            Note that surface data has not been calculated for all materials in the Materials \
            Project database. See the `SurfacePropDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Robocrystallographer",
        "description": "Route providing a computed text description for a material following the \
            methodology discussed by [Ganose *et al.*](https://doi.org/10.1557/mrc.2019.94) \
            Note that descriptions may not been calculated for all materials in the Materials \
            Project database. See the `RobocrysDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Synthesis",
        "description": "Route providing a synthesis recipes for materials extracted from literature \
            following the methodology discussed by [Kononova *et al.*](https://doi.org/10.1038/s41597-019-0224-1) \
            Note that synthesis recipes may not be available for all materials in the Materials \
            Project database. See the `SynthesisSearchResultModel` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Electrodes",
        "description": "Route providing computed electrode data for a material following the \
            methodology discussed by [Shen *et al.*](https://doi.org/10.1038/s41524-020-00422-3) \
            Note that electrode data has not been calculated for all materials in the Materials \
            Project database. See the `InsertionElectrodeDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Molecules",
        "description": "Route providing computed data for a molecule such as charge, \
            electron affinity, and ionization energy. The unique identifier for a molecule \
            is its `task_id` (e.g. `mol-45807`). See the `MoleculesDoc` schema for a full list of \
            fields returned by this route.",
    },
    {
        "name": "Oxidation States",
        "description": "Route providing computed oxidation state data for a material following the \
            methodology employed by the [BVAnalyzer](https://pymatgen.org/pymatgen.analysis.bond_valence.html) \
            in Pymatgen. Note that oxidation state data has not been calculated for all materials in the Materials \
            Project database. See the `OxidationStateDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "Provenance",
        "description": "Route providing provenance data for a material such as whether it is theoretical, \
            its associated ICSD entries, and relevant references in literature. Note that provenance data \
            may not be available for all materials in the Materials Project database. See the `ProvenanceDoc` \
            schema for a full list of fields returned by this route.",
    },
    {
        "name": "Charge Density",
        "description": "Route providing computed charge density data for a material following the \
            methodology discussed by [Shen *et al.*](https://arxiv.org/abs/2107.03540). Please email \
            <heavy.api.use@materialsproject.org> if you would like to retrieve a large amount of this data. \
            Note that charge densities may not be calculated for all materials in the Materials \
            Project database. See the `ChgcarDataDoc` schema for a full list of fields returned by this route.",
    },
    {
        "name": "MPComplete",
        "description": "Route for submitting structures to the Materials Project. If calculations are run with the \
            submitted structure, the submitter will be credited with the submitted public name and email.",
    },
    {
        "name": "Alloys",
        "description": "Route for retrevial of information about which hypothetical alloy(s) a given material might "
                       "belong to, following the methodolgy discussed by "
                       "[Woods-Robinson, Horton and Persson](https://arxiv.org/pdf/2206.10715)."
    }
]



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS

from collections import defaultdict
from pymatgen.core.composition import Composition

from emmet.core.grain_boundary import GBTypeEnum


class GBStructureQuery(QueryOperator):
    """
    Method to generate a query for structure related data associated with grain boundary entries
    """

    def query(
        self,
        sigma: Optional[int] = Query(None, description="Value of sigma.",),
        type: Optional[GBTypeEnum] = Query(None, description="Grain boundary type.",),
        chemsys: Optional[str] = Query(
            None, description="Dash-delimited string of elements in the material.",
        ),
        pretty_formula: Optional[str] = Query(
            None, description="Formula of the material.",
        ),
        gb_plane: Optional[str] = Query(
            None,
            description="Miller index of the grain boundary plane as comma delimitd integers.",
        ),
        rotation_axis: Optional[str] = Query(
            None,
            description="Miller index of the rotation axis as comma delimitd integers.",
        ),
    ) -> STORE_PARAMS:

        crit = defaultdict(dict)  # type: dict

        if sigma:
            crit["sigma"] = int(sigma)

        if type:
            crit["type"] = type.value

        if chemsys:
            chemsys = "-".join(sorted(chemsys.split("-")))
            crit["chemsys"] = chemsys

        if pretty_formula:
            crit["pretty_formula"] = Composition(pretty_formula).reduced_formula

        if gb_plane:
            crit["gb_plane"] = [int(n.strip()) for n in gb_plane.split(",")]

        if rotation_axis:
            crit["rotation_axis"] = [int(n.strip()) for n in rotation_axis.split(",")]

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = [key for key in self._keys_from_query() if "_min" not in key]
        keys.append("rotation_angle")
        return [(key, False) for key in keys]


class GBTaskIDQuery(QueryOperator):
    """
    Method to generate a query for different task_ids
    """

    def query(
        self,
        task_ids: Optional[str] = Query(
            None,
            description="Comma-separated list of Materials Project IDs to query on.",
        ),
    ) -> STORE_PARAMS:

        crit = {}

        if task_ids:
            crit.update(
                {
                    "task_id": {
                        "$in": [task_id.strip() for task_id in task_ids.split(",")]
                    }
                }
            )

        return {"criteria": crit}
